r"""
CT_NCM
##################################
Reference:
    Haiping Ma et al. "Reconciling cognitive modeling with knowledge forgetting: A continuous time-aware neural network approach." in IJCAI 2022.
Reference code:
    https://github.com/BIMK/Intelligent-Education/tree/main/CTNCM
"""
import torch
import torch.nn as nn
import torch.nn.functional as F
from ..gd_basemodel import GDBaseModel


class CT_NCM(GDBaseModel):
    """
    hidden_size: dimensions of LSTM hidden layerd
    embed_size: dimensions of student-knowledge concept interaction embedding
    prelen1: the first layer of performance prediction
    prelen2: the second layer of performance prediction
    dropout1: the proportion of first fully connected layer dropout before getting the prediction score
    dropout2: the proportion of second fully connected layer dropout before getting the prediction score
    """
    default_cfg = {
        'hidden_size': 64,
        'embed_size': 64,
        'prelen1': 256,
        'prelen2': 128,
        'dropout1': 0,
        'dropout2': 0,
    }

    def __init__(self, cfg):
        """Pass parameters from other templates into the model

        Args:
            cfg (UnifyConfig): parameters from other templates
        """
        super().__init__(cfg)

    def build_cfg(self):
        """Initialize the parameters of the model"""
        self.problem_num = self.datatpl_cfg['dt_info']['exer_count']
        self.skill_num = self.datatpl_cfg['dt_info']['cpt_count']
        self.device = self.traintpl_cfg['device']
        self.hidden_size = self.modeltpl_cfg['hidden_size']
        self.embed_size = self.modeltpl_cfg['embed_size']
        self.knowledge_dim = self.hidden_size
        self.input_len = self.knowledge_dim
        self.prelen1 = self.modeltpl_cfg['prelen1']
        self.prelen2 = self.modeltpl_cfg['prelen2']
        # 使用 BCEWithLogitsLoss，因为模型输出未经过 Sigmoid 激活
        self.loss_function = torch.nn.BCEWithLogitsLoss()
        # 如果选择使用 BCELoss，请取消上面一行并启用下面一行
        # self.loss_function = torch.nn.BCELoss()

    def build_model(self):
        """Initialize the various components of the model"""
        self.dropout1 = nn.Dropout(p=self.modeltpl_cfg['dropout1'])
        self.dropout2 = nn.Dropout(p=self.modeltpl_cfg['dropout2'])

        self.inter_embedding = torch.nn.Embedding(2 * self.skill_num, self.embed_size)
        self.reclstm = torch.nn.Linear(self.embed_size + self.hidden_size, 7 * self.hidden_size)

        self.problem_disc = torch.nn.Embedding(self.problem_num, 1)
        self.problem_diff = torch.nn.Embedding(self.problem_num, self.knowledge_dim)

        self.linear1 = torch.nn.Linear(self.input_len, self.prelen1)
        self.linear2 = torch.nn.Linear(self.prelen1, self.prelen2)
        self.linear3 = torch.nn.Linear(self.prelen2, 1)

        # 手动初始化权重
        nn.init.xavier_normal_(self.reclstm.weight)
        nn.init.xavier_normal_(self.linear1.weight)
        nn.init.xavier_normal_(self.linear2.weight)
        nn.init.xavier_normal_(self.linear3.weight)

    def forward(self, exer_seq, start_timestamp_seq, cpt_unfold_seq, label_seq, mask_seq, **kwargs):
        """A function of how well the model predicts students' responses to exercise questions

        Args:
            exer_seq (torch.Tensor): Sequence of exercise id. Shape of [batch_size, seq_len]
            start_timestamp_seq (torch.Tensor): Sequence of Students start answering time. Shape of [batch_size, seq_len]
            cpt_unfold_seq (torch.Tensor): Sequence of knowledge concepts related to exercises. Shape of [batch_size, seq_len]
            label_seq (torch.Tensor): Sequence of students' answers to exercises. Shape of [batch_size, seq_len]
            mask_seq (torch.Tensor): Sequence of mask. Mask=1 indicates that the student has answered the exercise, otherwise vice versa. Shape of [batch_size, seq_len]

        Returns:
            dict: The predictions of the model and the real situation
        """
        # 检查输入数据中是否存在 NaN 或 Inf
        assert not torch.isnan(exer_seq).any(), "exer_seq contains NaNs"
        assert not torch.isnan(start_timestamp_seq).any(), "start_timestamp_seq contains NaNs"
        assert not torch.isnan(cpt_unfold_seq).any(), "cpt_unfold_seq contains NaNs"
        assert not torch.isnan(label_seq).any(), "label_seq contains NaNs"
        assert not torch.isnan(mask_seq).any(), "mask_seq contains NaNs"
        assert not torch.isinf(exer_seq).any(), "exer_seq contains Inf"
        assert not torch.isinf(start_timestamp_seq).any(), "start_timestamp_seq contains Inf"
        assert not torch.isinf(cpt_unfold_seq).any(), "cpt_unfold_seq contains Inf"
        assert not torch.isinf(label_seq).any(), "label_seq contains Inf"
        assert not torch.isinf(mask_seq).any(), "mask_seq contains Inf"

        problem_seqs_tensor = exer_seq[:, 1:].to(self.device)
        skill_seqs_tensor = cpt_unfold_seq.to(self.device)
        start_timestamp_seqs_tensor = start_timestamp_seq[:, 1:].to(self.device)
        correct_seqs_tensor = label_seq.to(self.device)
        mask_labels = mask_seq.long().to(self.device)
        seqs_length = torch.sum(mask_labels, dim=1)
        delete_row = 0
        for i in range(len(seqs_length)):
            if seqs_length[i] == 1:
                mask = torch.arange(problem_seqs_tensor.size(0)) != (i - delete_row)
                problem_seqs_tensor = problem_seqs_tensor[mask]
                skill_seqs_tensor = skill_seqs_tensor[mask]
                start_timestamp_seqs_tensor = start_timestamp_seqs_tensor[mask]
                correct_seqs_tensor = correct_seqs_tensor[mask]
                mask_labels = mask_labels[mask]
                delete_row += 1

        # 将 mask_labels == 0 的标签设置为 0 而不是 -1
        correct_seqs_tensor = torch.where(mask_labels == 0, 0, correct_seqs_tensor)
        skill_seqs_tensor = torch.where(mask_labels == 0, 0, skill_seqs_tensor)
        mask_labels_temp = mask_labels[:, 1:]
        start_timestamp_seqs_tensor = torch.where(mask_labels_temp == 0, 0, start_timestamp_seqs_tensor)
        problem_seqs_tensor = torch.where(mask_labels_temp == 0, 0, problem_seqs_tensor)
        seqs_length = torch.sum(mask_labels, dim=1)

        # 再次检查处理后的数据
        assert not torch.isnan(problem_seqs_tensor).any(), "problem_seqs_tensor contains NaNs after processing"
        assert not torch.isnan(skill_seqs_tensor).any(), "skill_seqs_tensor contains NaNs after processing"
        assert not torch.isnan(start_timestamp_seqs_tensor).any(), "start_timestamp_seqs_tensor contains NaNs after processing"
        assert not torch.isnan(correct_seqs_tensor).any(), "correct_seqs_tensor contains NaNs after processing"

        inter_embed_tensor = self.inter_embedding(skill_seqs_tensor + self.skill_num * mask_labels)
        batch_size = correct_seqs_tensor.size()[0]

        hidden, _ = self.continues_lstm(inter_embed_tensor, start_timestamp_seqs_tensor, seqs_length, batch_size)
        hidden_packed = torch.nn.utils.rnn.pack_padded_sequence(hidden[1:, ],
                                                                seqs_length.cpu() - 1,
                                                                batch_first=False,
                                                                enforce_sorted=False)
        theta = hidden_packed.data
        problem_packed = torch.nn.utils.rnn.pack_padded_sequence(problem_seqs_tensor,
                                                                seqs_length.cpu() - 1,
                                                                batch_first=True,
                                                                enforce_sorted=False)
        predictions = torch.squeeze(self.problem_hidden(theta, problem_packed.data))
        labels_packed = torch.nn.utils.rnn.pack_padded_sequence(correct_seqs_tensor[:, 1:],
                                                                seqs_length.cpu() - 1,
                                                                batch_first=True,
                                                                enforce_sorted=False)
        labels = labels_packed.data
        out_dict = {'predictions': predictions, 'labels': labels}

        # 检查模型输出是否包含 NaN 或 Inf
        if torch.isnan(predictions).any():
            print("Predictions contain NaNs")
        if torch.isnan(labels).any():
            print("Labels contain NaNs")
        if torch.isinf(predictions).any():
            print("Predictions contain Inf")
        if torch.isinf(labels).any():
            print("Labels contain Inf")

        assert not torch.isnan(predictions).any(), "Predictions contain NaNs"
        assert not torch.isnan(labels).any(), "Labels contain NaNs"
        assert not torch.isinf(predictions).any(), "Predictions contain Inf"
        assert not torch.isinf(labels).any(), "Labels contain Inf"

        return out_dict

    def continues_lstm(self, inter_embed_tensor, start_timestamp_seqs_tensor, seqs_length, batch_size):
        """

        Args:
            inter_embed_tensor (torch.Tensor): interrelated LSTM unit. Shape of [batch_size, seq_len, embed_size]
            start_timestamp_seqs_tensor (torch.Tensor): Sequence of Students start answering time. Shape of [batch_size, seq_len-1]
            seqs_length (torch.Tensor): Length of sequence. Shape of [batch_size]
            batch_size (int): batch size.

        Returns:
            torch.Tensor: Output of LSTM.
        """
        self.init_states(batch_size=batch_size)
        h_list = [self.h_delay]
        for t in range(max(seqs_length) - 1):
            one_batch = inter_embed_tensor[:, t]
            c, self.c_bar, output_t, delay_t = self.conti_lstm(one_batch, self.h_delay, self.c_delay, self.c_bar)
            time_lag_batch = start_timestamp_seqs_tensor[:, t]
            self.c_delay, self.h_delay = self.delay(c, self.c_bar, output_t, delay_t, time_lag_batch)
            # 确保 h_delay 没有 NaN 或 Inf
            self.h_delay = torch.as_tensor(self.h_delay, dtype=torch.float).to(self.device)
            assert not torch.isnan(self.h_delay).any(), f"h_delay at time {t} contains NaNs"
            assert not torch.isinf(self.h_delay).any(), f"h_delay at time {t} contains Inf"
            h_list.append(self.h_delay)
        hidden = torch.stack(h_list)

        return hidden, seqs_length

    def init_states(self, batch_size):
        """Initialize the state of lstm

        Args:
            batch_size (int): batch_size
        """
        self.h_delay = torch.full((batch_size, self.hidden_size), 0.5, dtype=torch.float).to(self.device)
        self.c_delay = torch.full((batch_size, self.hidden_size), 0.5, dtype=torch.float).to(self.device)
        self.c_bar = torch.full((batch_size, self.hidden_size), 0.5, dtype=torch.float).to(self.device)
        self.c = torch.full((batch_size, self.hidden_size), 0.5, dtype=torch.float).to(self.device)

    def conti_lstm(self, one_batch_inter_embed, h_d_t, c_d_t, c_bar_t):
        """

        Args:
            one_batch_inter_embed (torch.Tensor): one batch of interrelated LSTM unit. Shape of [batch_size, embed_size]
            h_d_t (torch.Tensor): Shape of [batch_size, embed_size]
            c_d_t (torch.Tensor): Shape of [batch_size, embed_size]
            c_bar_t (torch.Tensor): Shape of [batch_size, embed_size]

        Returns:
            torch.Tensor: Data inside LSTM
        """
        input = torch.cat((one_batch_inter_embed, h_d_t), dim=1)
        (i, f, z, o, i_bar, f_bar, delay) = torch.chunk(self.reclstm(input), 7, -1)
        i = torch.sigmoid(i)
        f = torch.sigmoid(f)
        z = torch.tanh(z)
        o = torch.sigmoid(o)
        i_bar = torch.sigmoid(i_bar)
        f_bar = torch.sigmoid(f_bar)
        delay = F.softplus(delay)

        # 检查中间变量是否包含 NaN 或 Inf
        assert not torch.isnan(i).any(), "i contains NaNs"
        assert not torch.isnan(f).any(), "f contains NaNs"
        assert not torch.isnan(z).any(), "z contains NaNs"
        assert not torch.isnan(o).any(), "o contains NaNs"
        assert not torch.isnan(i_bar).any(), "i_bar contains NaNs"
        assert not torch.isnan(f_bar).any(), "f_bar contains NaNs"
        assert not torch.isnan(delay).any(), "delay contains NaNs"
        assert not torch.isinf(i).any(), "i contains Inf"
        assert not torch.isinf(f).any(), "f contains Inf"
        assert not torch.isinf(z).any(), "z contains Inf"
        assert not torch.isinf(o).any(), "o contains Inf"
        assert not torch.isinf(i_bar).any(), "i_bar contains Inf"
        assert not torch.isinf(f_bar).any(), "f_bar contains Inf"
        assert not torch.isinf(delay).any(), "delay contains Inf"

        c_t = f * c_d_t + i * z
        c_bar_t = f_bar * c_bar_t + i_bar * z

        # 检查 c_t 和 c_bar_t 是否包含 NaN 或 Inf
        assert not torch.isnan(c_t).any(), "c_t contains NaNs"
        assert not torch.isnan(c_bar_t).any(), "c_bar_t contains NaNs"
        assert not torch.isinf(c_t).any(), "c_t contains Inf"
        assert not torch.isinf(c_bar_t).any(), "c_bar_t contains Inf"

        return c_t, c_bar_t, o, delay

    def delay(self, c, c_bar, output, delay, time_lag):
        """
        Args:
            c (torch.Tensor): Shape of [batch_size, embed_size]
            c_bar (torch.Tensor): Shape of [batch_size, embed_size]
            output (torch.Tensor): Shape of [batch_size, embed_size]
            delay (torch.Tensor): Shape of [batch_size, embed_size]
            time_lag (torch.Tensor): Shape of [batch_size]

        Returns:
            torch.Tensor: Data inside LSTM
        """
        exponent = - delay * time_lag.unsqueeze(-1)
        exponent = torch.clamp(exponent, min=-20, max=10)  # 限制指数的范围，max 降低到 10
        delta = c - c_bar
        delta = torch.clamp(delta, min=-10, max=10)  # 限制 delta 的范围
        c_delay = c_bar + delta * torch.exp(exponent)
        h_delay = output * torch.tanh(c_delay)

        # 检查 c_delay 和 h_delay 是否包含 NaN 或 Inf
        assert not torch.isnan(c_delay).any(), "c_delay contains NaNs"
        assert not torch.isnan(h_delay).any(), "h_delay contains NaNs"
        assert not torch.isinf(c_delay).any(), "c_delay contains Inf"
        assert not torch.isinf(h_delay).any(), "h_delay contains Inf"

        return c_delay, h_delay

    def problem_hidden(self, theta, problem_data):
        """Get how well the model predicts students' responses to exercise questions

        Args:
            theta (torch.Tensor): Student's ability value. Shape of [exer_num, seq_len]
            problem_data (torch.Tensor): The id of the exercise that the student has answered. Shape of [exer_num]

        Returns:
            torch.Tensor: the model predictions of students' responses to exercise questions. Shape of [exer_num, 1]
        """
        problem_diff = torch.sigmoid(self.problem_diff(problem_data))
        problem_disc = torch.sigmoid(self.problem_disc(problem_data))
        input_x = (theta - problem_diff) * problem_disc  # 移除 * 10

        # 添加限制
        input_x = torch.clamp(input_x, min=-10, max=10)  # 根据数据情况调整范围

        # 检查 input_x 是否包含 NaN 或 Inf
        assert not torch.isnan(input_x).any(), "input_x contains NaNs"
        assert not torch.isinf(input_x).any(), "input_x contains Inf"

        # 使用 ReLU 替代 Sigmoid
        input_x = self.dropout1(F.relu(self.linear1(input_x)))

        # 检查 input_x 是否包含 NaN 或 Inf
        assert not torch.isnan(input_x).any(), "input_x after linear1 and ReLU contains NaNs"
        assert not torch.isinf(input_x).any(), "input_x after linear1 and ReLU contains Inf"

        # 使用 ReLU 替代 Sigmoid
        input_x = self.dropout2(F.relu(self.linear2(input_x)))

        # 检查 input_x 是否包含 NaN 或 Inf
        assert not torch.isnan(input_x).any(), "input_x after linear2 and ReLU contains NaNs"
        assert not torch.isinf(input_x).any(), "input_x after linear2 and ReLU contains Inf"

        output = self.linear3(input_x)  # 移除 Sigmoid 激活

        # 检查 output 是否包含 NaN 或 Inf
        assert not torch.isnan(output).any(), "output contains NaNs"
        assert not torch.isinf(output).any(), "output contains Inf"

        # 可选：打印 output 的统计信息
        # print(f"Output - min: {output.min().item()}, max: {output.max().item()}")

        return output

    def predict(self, **kwargs):
        """A function of get how well the model predicts students' responses to exercise questions and the groundtruth

        Returns:
            dict: The predictions of the model and the real situation
        """
        outdict = self(**kwargs)
        # 在预测阶段应用 Sigmoid 激活以获取概率值
        y_pd = torch.sigmoid(outdict['predictions'])
        return {
            'y_pd': y_pd,
            'y_gt': torch.as_tensor(outdict['labels'], dtype=torch.float)
        }

    def get_main_loss(self, **kwargs):
        """

        Returns:
            dict: loss dict{'loss_main': loss_value}
        """
        outdict = self(**kwargs)
        predictions = outdict['predictions']
        labels = outdict['labels']
        labels = torch.as_tensor(labels, dtype=torch.float)

        # 添加断言以确保标签在 [0, 1] 范围内
        assert torch.all((labels == 0) | (labels == 1)), "标签中存在非 0 或 1 的值"

        # 检查 predictions 是否包含 NaN 或 Inf
        assert not torch.isnan(predictions).any(), "Predictions contain NaNs"
        assert not torch.isinf(predictions).any(), "Predictions contain Inf"

        loss = self.loss_function(predictions, labels)
        return {
            'loss_main': loss
        }

    def get_loss_dict(self, **kwargs):
        """

        Returns:
            dict: loss dict{'loss_main': loss_value}
        """
        return self.get_main_loss(**kwargs)



from ..common.base_mid2cache import BaseMid2Cache
import numpy as np
from edustudio.datatpl.utils import PadSeqUtil
import pandas as pd


class M2C_GenUnFoldKCSeq(BaseMid2Cache):
    default_cfg = {}
    
    def __init__(self, m2c_cfg, n_folds, is_dataset_divided) -> None:
        super().__init__(m2c_cfg)
        self.n_folds = n_folds
        self.is_dataset_divided = is_dataset_divided

    @classmethod
    def from_cfg(cls, cfg):
        m2c_cfg = cfg.datatpl_cfg.get(cls.__name__)
        n_folds = cfg.datatpl_cfg.n_folds
        is_dataset_divided = cfg.datatpl_cfg.is_dataset_divided
        return cls(m2c_cfg, n_folds, is_dataset_divided)
    
    def process(self, **kwargs):
        df = kwargs['df']
        df_exer = kwargs['df_exer']
        df_train, df_valid, df_test = kwargs['df_train'], kwargs['df_valid'], kwargs['df_test']


        if not self.is_dataset_divided:
            assert df_train is None and df_valid is None and df_test is None
            unique_cpt_seq = df_exer['cpt_seq:token_seq'].explode().unique()
            cpt_map = dict(zip(unique_cpt_seq, range(len(unique_cpt_seq))))
            df_Q_unfold = pd.DataFrame({
                'exer_id:token': df_exer['exer_id:token'].repeat(df_exer['cpt_seq:token_seq'].apply(len)),
                'cpt_unfold:token': df_exer['cpt_seq:token_seq'].explode().replace(cpt_map)
            })
            df = pd.merge(df, df_Q_unfold, on=['exer_id:token'], how='left').reset_index(drop=True)
            kwargs['df'] = df
        else: # dataset is divided
            assert df_train is not None and df_test is not None
            train_df, val_df, test_df = self._unfold_dataset(df_train, df_valid, df_test, df_exer)
            kwargs['df_train'] = train_df
            kwargs['df_valid'] = val_df
            kwargs['df_test'] = test_df
        return kwargs

    def _unfold_dataset(self, train_df, val_df, test_df, df_Q):
        unique_cpt_seq = df_Q['cpt_seq'].explode().unique()
        cpt_map = dict(zip(unique_cpt_seq, range(len(unique_cpt_seq))))
        df_Q_unfold = pd.DataFrame({
            'exer_id:token': df_Q['exer_id'].repeat(df_Q['cpt_seq'].apply(len)),
            'cpt_unfold:token': df_Q['cpt_seq'].explode().replace(cpt_map)
        })
        train_df_unfold = pd.merge(train_df, df_Q_unfold, on=['exer_id'], how='left').reset_index(drop=True)
        val_df_unfold = pd.merge(val_df, df_Q_unfold, on=['exer_id'], how='left').reset_index(drop=True)
        test_df_unfold = pd.merge(test_df, df_Q_unfold, on=['exer_id'], how='left').reset_index(drop=True)

        return train_df_unfold, val_df_unfold, test_df_unfold



r"""
CT_NCM
##################################
Reference:
    Chenyang Wang et al. "Temporal cross-effects in knowledge tracing." in WSDM 2021.
"""
import numpy as np
from ..gd_basemodel import GDBaseModel
import torch
import torch.nn.functional as F


class HawkesKT(GDBaseModel):
    default_cfg = {
        'dim_s': 50,  # 序列长度
        'emb_size': 64,
        'time_log': 5,  # Log base of time intervals.
    }

    def __init__(self, cfg):
        """Pass parameters from other templates into the model

        Args:
            cfg (UnifyConfig): parameters from other templates
        """
        super().__init__(cfg)

    def build_cfg(self):
        """Initialize the parameters of the model"""
        self.problem_num = self.datatpl_cfg['dt_info']['exer_count']
        self.skill_num = self.datatpl_cfg['dt_info']['cpt_count']
        self.emb_size = self.modeltpl_cfg['emb_size']
        self.time_log = self.modeltpl_cfg['time_log']

    def build_model(self):
        """Initialize the various components of the model"""
        self.problem_base = torch.nn.Embedding(self.problem_num, 1)
        self.skill_base = torch.nn.Embedding(self.skill_num, 1)

        self.alpha_inter_embeddings = torch.nn.Embedding(self.skill_num * 2, self.emb_size)  # Corresponding to the PA in the paper
        self.alpha_skill_embeddings = torch.nn.Embedding(self.skill_num, self.emb_size)
        self.beta_inter_embeddings = torch.nn.Embedding(self.skill_num * 2, self.emb_size)
        self.beta_skill_embeddings = torch.nn.Embedding(self.skill_num, self.emb_size)

    def _init_params(self):
        """Parameter initialization of each component of the model"""
        torch.nn.init.normal_(self.problem_base.weight, mean=0.0, std=0.01)
        torch.nn.init.normal_(self.skill_base.weight, mean=0.0, std=0.01)
        torch.nn.init.normal_(self.alpha_inter_embeddings.weight, mean=0.0, std=0.01)
        torch.nn.init.normal_(self.alpha_skill_embeddings.weight, mean=0.0, std=0.01)
        torch.nn.init.normal_(self.beta_inter_embeddings.weight, mean=0.0, std=0.01)
        torch.nn.init.normal_(self.beta_skill_embeddings.weight, mean=0.0, std=0.01)

    def forward(self, exer_seq, start_timestamp_seq, cpt_unfold_seq, **kwargs):
        """A function of how well the model predicts students' responses to exercise questions

        Args:
            exer_seq (torch.Tensor): Sequence of exercise id. Shape of [batch_size, seq_len]
            start_timestamp_seq (torch.Tensor): The time the student started answering the question. Shape of [batch_size, seq_len]
            cpt_unfold_seq (torch.Tensor): Knowledge concepts corresponding to exercises. Shape of [batch_size, seq_len]

        Returns:
            torch.Tensor: The predictions of the model
        """
        skills = cpt_unfold_seq     # [batch_size, seq_len] One exercise corresponds to one knowledge point
        problems = exer_seq  # [batch_size, seq_len] sequence of batch_size students
        # time = [i for i in range(start_timestamp_seq.shape[1])]
        # times = torch.Tensor([time for i in range(start_timestamp_seq.shape[0])])
        times = start_timestamp_seq - start_timestamp_seq[:,[0]]        # [batch_size, seq_len]

        mask_labels = kwargs['mask_seq'].long()
        inters = skills + mask_labels * self.skill_num

        alpha_src_emb = self.alpha_inter_embeddings(inters)  # [bs, seq_len, emb]
        alpha_target_emb = self.alpha_skill_embeddings(skills)
        alphas = torch.matmul(alpha_src_emb, alpha_target_emb.transpose(-2, -1))  # [bs, seq_len, seq_len]
        beta_src_emb = self.beta_inter_embeddings(inters)  # [bs, seq_len, emb]
        beta_target_emb = self.beta_skill_embeddings(skills)
        betas = torch.matmul(beta_src_emb, beta_target_emb.transpose(-2, -1))  # [bs, seq_len, seq_len]
        betas = torch.clamp(betas + 1, min=0, max=10)  

        delta_t = (times[:, :, None] - times[:, None, :]).abs().double()  # Get the absolute value of the time at different time steps
        delta_t = torch.log(delta_t + 1e-10) / np.log(self.time_log)

        cross_effects = alphas * torch.exp(-betas * delta_t)  # The cross_effects of the paper (4)

        seq_len = skills.shape[1]
        valid_mask = np.triu(np.ones((1, seq_len, seq_len)), k=1)
        mask = (torch.from_numpy(valid_mask) == 0)
        mask = mask.cuda() if self.device != 'cpu' else mask
        sum_t = cross_effects.masked_fill(mask, 0).sum(-2)

        problem_bias = self.problem_base(problems).squeeze(dim=-1)
        skill_bias = self.skill_base(skills).squeeze(dim=-1)

        prediction = (problem_bias + skill_bias + sum_t).sigmoid()
        return prediction

    @torch.no_grad()
    def predict(self, **kwargs):
        """A function of get how well the model predicts students' responses to exercise questions and the groundtruth

        Returns:
            dict: The predictions of the model and the real situation
        """
        y_pd = self(**kwargs)
        y_pd = y_pd[:, :-1]
        y_pd = y_pd[kwargs['mask_seq'][:, 1:] == 1]
        y_gt = None
        if kwargs.get('label_seq', None) is not None:
            y_gt = kwargs['label_seq'][:, 1:]
            y_gt = y_gt[kwargs['mask_seq'][:, 1:] == 1]
        return {
            'y_pd': y_pd,
            'y_gt': y_gt
        }

    def get_main_loss(self, **kwargs):
        """

        Returns:
            dict: {'loss_main': loss_value}
        """
        y_pd = self(**kwargs)
        y_pd = y_pd[:, :-1]
        y_pd = y_pd[kwargs['mask_seq'][:, 1:] == 1]
        y_gt = kwargs['label_seq'][:, 1:]
        y_gt = y_gt[kwargs['mask_seq'][:, 1:] == 1]
        loss = F.binary_cross_entropy(
            input=y_pd.double(), target=y_gt.double()
        )
        return {
            'loss_main': loss
        }

    def get_loss_dict(self, **kwargs):
        """

        Returns:
            dict: {'loss_main': loss_value}
        """
        return self.get_main_loss(**kwargs)



from .KTInterDataTPL import KTInterDataTPL
from .KTInterExtendsQDataTPL import KTInterExtendsQDataTPL
from .KTInterCptAsExerDataTPL import KTInterCptAsExerDataTPL
from .EERNNDataTPL import EERNNDataTPL
from .LPKTDataTPL import LPKTDataTPL
from .DKTForgetDataTPL import DKTForgetDataTPL
from .KTInterCptUnfoldDataTPL import KTInterCptUnfoldDataTPL
from .DKTDSCDataTPL import DKTDSCDataTPL
from .DIMKTDataTPL import DIMKTDataTPL
from .QDKTDataTPL import QDKTDataTPL
from .RKTDataTPL import RKTDataTPL
from .CL4KTDataTPL import CL4KTDataTPL
from .EKTDataTPL import EKTDataTPL
from .GKTDataTPL import GKTDataTPL



from ..common import BaseMid2Cache
import torch

class M2C_CNCDQ_OP(BaseMid2Cache):
    default_cfg = {}

    def process(self, **kwargs):
        df_exer = kwargs['df_exer']
        exer_count = kwargs['dt_info']['exer_count']
        cpt_count = kwargs['dt_info']['cpt_count']
        kwargs['Q_mask_mat'], kwargs['knowledge_pairs'] = self._get_knowledge_pairs(
            df_Q_arr=df_exer, exer_count=exer_count, cpt_count=cpt_count
        )
        return kwargs

    def _get_knowledge_pairs(self, df_Q_arr, exer_count, cpt_count):
        Q_mask_mat = torch.zeros((exer_count, cpt_count), dtype=torch.int64)
        knowledge_pairs = []
        kn_tags = []
        kn_topks = []
        for _, item in df_Q_arr.iterrows():
            # kn_tags.append(item['cpt_seq'])
            # kn_topks.append(item['cpt_pre_seq'])
            kn_tags = item['cpt_seq']
            kn_topks = item['cpt_pre_seq']
            knowledge_pairs.append((kn_tags, kn_topks))
            for cpt_id in item['cpt_seq']:
                Q_mask_mat[item['exer_id'], cpt_id-1] = 1
            for cpt_id in item['cpt_pre_seq']:
                Q_mask_mat[item['exer_id'], cpt_id-1] = 1
        return Q_mask_mat, knowledge_pairs



import pandas as pd
from .general_datatpl import DataTPLStatus
from edustudio.utils.common import UnifyConfig
from .general_datatpl import GeneralDataTPL
import os
import torch


class EduDataTPL(GeneralDataTPL):
    """Educational Data Template including student features and exercise features
    """
    
    default_cfg = {
        'exer_exclude_feat_names': (),
        'stu_exclude_feat_names': (),
    }

    def __init__(
                self, cfg:UnifyConfig,
                df: pd.DataFrame=None,
                df_train: pd.DataFrame=None,
                df_valid: pd.DataFrame=None,
                df_test: pd.DataFrame=None,
                df_stu: pd.DataFrame=None,
                df_exer: pd.DataFrame=None,
                status: DataTPLStatus=DataTPLStatus()
            ):
        self.df_stu = df_stu
        self.df_exer = df_exer
        super().__init__(cfg, df, df_train, df_valid, df_test, status)
        assert self.df_stu is not None or self.df_exer is not None

    @property
    def common_str2df(self):
        return {
            "df": self.df, "df_train": self.df_train, "df_valid": self.df_valid,
            "df_test": self.df_test, "df_stu": self.df_stu, "df_exer": self.df_exer, "dt_info": self.datatpl_cfg['dt_info']
        }

    @classmethod
    def load_data(cls, cfg): # 只在middata存在时调用
        kwargs = super().load_data(cfg)
        new_kwargs = cls.load_data_from_side_information(cfg)
        for df in new_kwargs.values(): 
            if df is not None:
                cls._preprocess_feat(df) # 类型转换
        kwargs.update(new_kwargs)
        return kwargs
    
    @classmethod
    def load_data_from_side_information(cls, cfg):
        stu_fph = f'{cfg.frame_cfg.data_folder_path}/middata/{cfg.dataset}.stu.csv'
        exer_fph = f'{cfg.frame_cfg.data_folder_path}/middata/{cfg.dataset}.exer.csv'

        sep = cfg.datatpl_cfg['seperator']
        stu_exclude_feats = cfg.datatpl_cfg['stu_exclude_feat_names']
        exer_exclude_feats = cfg.datatpl_cfg['exer_exclude_feat_names']
        assert len(set(stu_exclude_feats) & {'stu_id:token'}) == 0
        assert len(set(exer_exclude_feats) & {'exer_id:token'}) == 0

        df_stu, df_exer = None, None
        if os.path.exists(stu_fph):
            df_stu = cls._load_atomic_csv(stu_fph, exclude_headers=stu_exclude_feats, sep=sep)

        if os.path.exists(exer_fph):
            df_exer = cls._load_atomic_csv(exer_fph, exclude_headers=exer_exclude_feats, sep=sep)

        return {"df_stu": df_stu, "df_exer": df_exer}
    
    @property
    def hasStuFeats(self):
        return self.df_stu is not None
    
    @property
    def hasExerFeats(self):
        return self.df_exer is not None

    @property
    def hasQmat(self):
        return self.hasExerFeats and "cpt_list" in self.df_exer.columns

    def save_cache(self):
        super().save_cache()
        df_stu_fph = f"{self.cache_folder_path}/df_stu.pkl"
        df_exer_fph = f"{self.cache_folder_path}/df_exer.pkl"
        self.save_pickle(df_stu_fph, self.df_stu)
        self.save_pickle(df_exer_fph, self.df_exer)
    
    def load_cache(self):
        super().load_cache()
        df_stu_fph = f"{self.cache_folder_path}/df_stu.pkl"
        df_exer_fph = f"{self.cache_folder_path}/df_exer.pkl"
        self.df_stu = self.load_pickle(df_stu_fph)
        self.df_exer = self.load_pickle(df_exer_fph)



r"""
ATKT
##################################
Reference:
    Xiaopeng Guo et al. "Enhancing knowledge tracing via adversarial training." in MM 2021.
Reference code:
    https://github.com/xiaopengguo/ATKT
"""

from ..gd_basemodel import GDBaseModel
import torch.nn as nn
import torch
import torch.nn.functional as F


class ATKT(GDBaseModel):
    """
    skill_dim: dimensions of knowledge concept representation
    answer_dim: representation dimensions of students' answers to exercises
    hidden_dim: dimensions of the hidden layer between the convolutional neural network and the output layer
    attention_dim: dimensions of parameters in the attention mechanism
    """
    default_cfg = {
        'skill_dim': 256,
        'answer_dim': 96,
        'hidden_dim': 80,
        'attention_dim': 80,
    }

    def __init__(self, cfg):
        """Pass parameters from other templates into the model

        Args:
            cfg (UnifyConfig): parameters from other templates
        """
        super().__init__(cfg)

    def _init_params(self):
        """Skip initialization of parameters for individual components of the model"""
        pass

    def build_cfg(self):
        """Initialize the parameters of the model"""
        self.output_dim = self.datatpl_cfg['dt_info']['cpt_count']
        self.skill_dim = self.modeltpl_cfg['skill_dim']
        self.answer_dim = self.modeltpl_cfg['answer_dim']
        self.hidden_dim = self.modeltpl_cfg['hidden_dim']
        self.attention_dim = self.modeltpl_cfg['attention_dim']

    def build_model(self):
        """Initialize the various components of the model"""
        self.rnn = nn.LSTM(self.skill_dim+self.answer_dim, self.hidden_dim, batch_first=True)
        self.fc = nn.Linear(self.hidden_dim*2, self.output_dim)
        self.sig = nn.Sigmoid()
        
        self.skill_emb = nn.Embedding(self.output_dim+1, self.skill_dim)
        self.skill_emb.weight.data[-1]= 0
        
        self.answer_emb = nn.Embedding(2+1, self.answer_dim)
        self.answer_emb.weight.data[-1]= 0

        self.mlp = nn.Linear(self.hidden_dim, self.attention_dim)
        self.similarity = nn.Linear(self.attention_dim, 1, bias=False)

    def _get_next_pred(self, res, skill):
        """Get how well the model predicts students' responses to exercise questions

        Args:
            res (torch.Tensor): Shape of [batch_size, seq_len-1, 123]
            skill (torch.Tensor): Sequence of knowledge concepts related to exercises. Shape of [batch_size, seq_len]

        Returns:
            torch.Tensor: the model predictions of students' responses to exercise questions. Shape of [batch_size, seq_len-1]
        """
        one_hot = torch.eye(self.output_dim, device=res.device)
        one_hot = torch.cat((one_hot, torch.zeros(1, self.output_dim).to(self.device)), dim=0)
        next_skill = skill[:, 1:]
        one_hot_skill = F.embedding(next_skill, one_hot)
        
        pred = (res * one_hot_skill).sum(dim=-1)
        return pred
    
    def attention_module(self, lstm_output):
        """

        Args:
            lstm_output (torch.Tensor): output of lstm. Shape of [batch_size, seq_len, attention_dim]

        Returns:
            torch.Tensor: output of attention module. Shape of [batch_size, seq_len, 2*attention_dim]
        """
        att_w = self.mlp(lstm_output)
        att_w = torch.tanh(att_w)
        att_w = self.similarity(att_w)
        
        alphas=nn.Softmax(dim=1)(att_w)
        
        attn_ouput=alphas*lstm_output
        attn_output_cum=torch.cumsum(attn_ouput, dim=1)
        attn_output_cum_1=attn_output_cum-attn_ouput

        final_output=torch.cat((attn_output_cum_1, lstm_output),2)
        
        return final_output

    def forward(self, cpt_unfold_seq, label_seq, mask_seq, p_adv=None, **kwargs):
        """A function of how well the model predicts students' responses to exercise questions

        Args:
            cpt_unfold_seq (torch.Tensor): Sequence of knowledge concepts related to exercises. Shape of [batch_size, seq_len]
            label_seq (torch.Tensor): Sequence of students' answers to exercises. Shape of [batch_size, seq_len]
            mask_seq (torch.Tensor): Sequence of mask. Mask=1 indicates that the student has answered the exercise, otherwise vice versa. Shape of [batch_size, seq_len] 
            p_adv (torch.Tensor, optional): perturbation to skill_answer_embedding. Defaults to None.

        Returns:
            torch.Tensor: The predictions of the model and the skill answer embedding
        """
        skill = torch.where(mask_seq==0, torch.tensor([self.output_dim]).to(self.device), cpt_unfold_seq)
        answer = torch.where(mask_seq==0, torch.tensor([2]).to(self.device), label_seq).long()
        perturbation = p_adv
        
        skill_embedding=self.skill_emb(skill)  # skill:24*500(batch_size*seq_len)
        answer_embedding=self.answer_emb(answer)  # answer:24*500
        
        skill_answer=torch.cat((skill_embedding,answer_embedding), 2)
        answer_skill=torch.cat((answer_embedding,skill_embedding), 2)
        
        answer=answer.unsqueeze(2).expand_as(skill_answer)
        
        skill_answer_embedding=torch.where(answer==1, skill_answer, answer_skill)

        skill_answer_embedding1=skill_answer_embedding
        
        if  perturbation is not None:
            skill_answer_embedding+=perturbation
        
        out,_ = self.rnn(skill_answer_embedding)
        out=self.attention_module(out)
        res = self.sig(self.fc(out))

        res = res[:, :-1, :]
        pred_res = self._get_next_pred(res, skill)  # 24*499(batch_size*(seq_len-1))
        
        return pred_res, skill_answer_embedding1

    @torch.no_grad()
    def predict(self, **kwargs):
        """A function of get how well the model predicts students' responses to exercise questions and the groundtruth

        Returns:
            dict: The predictions of the model and the real situation
        """
        y_pd, _ = self(**kwargs)
        y_pd = y_pd[kwargs['mask_seq'][:, 1:] == 1]
        y_gt = None
        if kwargs.get('label_seq', None) is not None:
            y_gt = kwargs['label_seq'][:, 1:]
            y_gt = y_gt[kwargs['mask_seq'][:, 1:] == 1]
        return {
            'y_pd': y_pd,
            'y_gt': y_gt
        }






import pandas as pd
import random
import os
from .raw2mid import BaseRaw2Mid

r"""
R2M_EdNet_KT1
########################
"""


class R2M_EdNet_KT1(BaseRaw2Mid):
    """R2M_EdNet_KT1 is a class used to handle the EdNet-KT1 dataset."""

    def process(self):
        super().process()
        pd.set_option("mode.chained_assignment", None)  # ignore warning
        pd.set_option("mode.chained_assignment", None)  # ignore warning
        # 读取数据集，并展示相关数据
        # 数据集包括784,309个学生，每个学生有一个交互的csv文件，这里我们进行抽样其中的5000个
        # 抽样其中的5000个
        all_files = pd.Series(os.listdir(f"{self.rawpath}")).to_list()
        all_files.remove("contents")  # 除去contents文件夹
        random.seed(2)
        files = random.sample(all_files, 5000)  # 采样个数5000个
        all_data = []
        for file_name in files:
            data = pd.read_csv(f"{self.rawpath}" + '\\' + file_name, encoding="utf-8")
            data['stu_id'] = int(file_name[:-4][1:])
            # 先把每个用户的数据暂存到列表[]中， 后一次性转化为DataFrame
            all_data.append(data)
        df = pd.concat(all_data)
        # 确定order字段
        df = df.sort_values(by='timestamp', ascending=True)
        df['order'] = range(len(df))

        # 读取question.csv,判断用户作答情况，是否正确
        question = pd.read_csv(os.path.join(f"{self.rawpath}", 'contents', 'questions.csv'))
        inter = df.merge(question, sort=False, how='left')
        inter = inter.dropna(subset=["stu_id", "question_id", "elapsed_time", "timestamp", "tags", "user_answer"])
        inter['label'] = (inter['correct_answer'] == inter['user_answer']).apply(int)
        inter['exer_id'] = inter['question_id'].apply(lambda x: x[1:])

        # 交互信息
        df_inter = inter \
            .reindex(columns=['stu_id', 'exer_id', 'label', 'timestamp', 'elapsed_time', 'order']) \
            .rename(columns={'stu_id': 'stu_id:token', 'exer_id': 'exer_id:token', 'label': 'label:float',
                             'timestamp': 'start_timestamp:float', 'elapsed_time': 'cost_time:float',
                             'order': 'order_id:float'})

        # 处理用户信息
        df_stu = df_inter['stu_id:token'].copy().unique()

        # 处理习题信息
        exer = question.copy()
        exer['cpt_seq'] = exer['tags'].apply(lambda x: x.split(';')).apply(lambda x: list(map(int, x)))
        exer['exer_id'] = exer['question_id'].apply(lambda x: x[1:])
        # 存储所有的cpt，即知识点id
        kcs = set()
        for cpt in exer['cpt_seq']:
            kcs.update(cpt)
        df_exer = exer.reindex(columns=['exer_id', 'cpt_seq']).rename(columns={
            'exer_id': 'exer_id:token', 'cpt_seq': 'cpt_seq:token_seq'
        })
        # 此处将数据保存到`cfg.MIDDATA_PATH`中
        df_inter.to_csv(f"{self.midpath}/{self.dt}.inter.csv", index=False, encoding='utf-8')
        # df_stu.to_csv(f"{cfg.MIDDATA_PATH}/{cfg.DT}.stu.csv", index=False, encoding='utf-8')
        df_exer.to_csv(f"{self.midpath}/{self.dt}.exer.csv", index=False, encoding='utf-8')
        pd.set_option("mode.chained_assignment", "warn")  # ignore warning
        return



from ..common.base_mid2cache import BaseMid2Cache
from sklearn.preprocessing import LabelEncoder
from itertools import chain


class M2C_KCAsExer(BaseMid2Cache):
    """Knowledge Concept As Exercise
    """
    default_cfg = {}

    def process(self, **kwargs):
        df = kwargs['df']
        df_exer = kwargs['df_exer']
        df_train, df_valid, df_test = kwargs['df_train'], kwargs['df_valid'], kwargs['df_test']

        if df is not None:
            assert df_train is None and df_valid is None and df_test is None
            kwargs['df_exer'], kwargs['df'] = self.cpt_as_exer_4undivided(df_exer, df)

        else:
            assert df_train is not None and df_test is not None
            kwargs['df_exer'], kwargs['df_train'], kwargs['df_valid'], kwargs['df_test'] = self.cpt_as_exer_4divided(df_exer, df_train, df_valid, df_test)
        
        return kwargs

    @staticmethod
    def cpt_as_exer_4undivided(df_exer, df):
        df_exer.rename(columns={"exer_id:token": "exer_id_old:token"}, inplace=True)
        df_exer['cpt_seq:token_seq'] = df_exer['cpt_seq:token_seq'].apply(lambda x: "_".join(list(map(lambda i: str(i), set(x)))))
        lbe = LabelEncoder()
        df_exer['cpt_seq:token_seq'] = lbe.fit_transform(df_exer['cpt_seq:token_seq']).tolist()
        df_exer.rename(columns={"cpt_seq:token_seq": "exer_id:token"}, inplace=True)

        df.rename(columns={'exer_id:token': 'exer_id_old:token'}, inplace=True)
        df = df.merge(df_exer[['exer_id_old:token', "exer_id:token"]], on='exer_id_old:token', how='left').reset_index(drop=True)
        df.drop(labels=['exer_id_old:token'], axis=1, inplace=True)
        return df_exer, df

    @staticmethod
    def cpt_as_exer_4divided(df_exer, df_train, df_valid, df_test):
        df_exer.rename(columns={"exer_id:token": "exer_id_old:token"}, inplace=True)
        df_exer['cpt_seq:token_seq'] = df_exer['cpt_seq:token_seq'].apply(
            lambda x: "_".join(list(map(lambda i: str(i), set(x))))
        )
        lbe = LabelEncoder()
        df_exer['cpt_seq:token_seq'] = lbe.fit_transform(df_exer['cpt_seq:token_seq']).tolist()
        df_exer.rename(columns={"cpt_seq:token_seq": "exer_id:token"}, inplace=True)

        df_train.rename(columns={'exer_id:token': 'exer_id_old:token'}, inplace=True)
        df_train = df_train.merge(df_exer, on='exer_id_old:token', how='left').reset_index(drop=True)
        df_train.drop(labels=['exer_id_old:token'], axis=1, inplace=True)
        if df_valid is not None:
            df_valid.rename(columns={'exer_id:token': 'exer_id_old:token'}, inplace=True)
            df_valid = df_valid.merge(df_exer, on='exer_id_old:token', how='left').reset_index(drop=True)
            df_valid.drop(labels=['exer_id_old:token'], axis=1, inplace=True)
        df_test.rename(columns={'exer_id:token': 'exer_id_old:token'}, inplace=True)
        df_test = df_test.merge(df_exer[['exer_id_old:token', "exer_id:token"]], on='exer_id_old:token', how='left').reset_index(drop=True)
        df_test.drop(labels=['exer_id_old:token'], axis=1, inplace=True)

        return df_exer, df_train, df_valid, df_test

    def set_dt_info(self, dt_info, **kwargs):
        if 'stu_id:token' in kwargs['df'].columns:
            dt_info['stu_count'] = int(kwargs['df']['stu_id:token'].max() + 1)
        if 'exer_id:token' in kwargs['df'].columns:
            dt_info['exer_count'] = int(kwargs['df']['exer_id:token'].max() + 1)
        if kwargs.get('df_exer', None) is not None:
            if 'cpt_seq:token_seq' in kwargs['df_exer']:
                dt_info['cpt_count'] = len(set(list(chain(*kwargs['df_exer']['cpt_seq:token_seq'].to_list()))))

    


# -*- coding: utf-8 -*-
# @Author : Xiangzhi Chen
# @Github : kervias

import copy
from typing import Union, Any, Optional
import os
import json
import yaml
import types
import re
from functools import reduce
import numpy as np


class UnifyConfig(object):
    """
        Unified config object
    """

    def __init__(self, dic: Optional[dict] = None):
        self.__config__ = dic or dict()

    @classmethod
    def from_py_module(cls, module_object: types.ModuleType):
        return cls(
            {k: (getattr(module_object, k) if not k.endswith("_PATH") else os.path.realpath(getattr(module_object, k)))
             for k in dir(module_object) if k[0].isupper()}
        )

    @classmethod
    def from_yml_file(cls, filepath: str):
        with open(filepath, 'r', encoding='utf-8') as f:
            config = yaml.load(f, Loader=cls._build_yaml_loader())
        return cls(config or dict())

    @staticmethod
    def _build_yaml_loader():
        loader = yaml.FullLoader
        loader.add_implicit_resolver(
            u'tag:yaml.org,2002:float',
            re.compile(
                u'''^(?:
             [-+]?(?:[0-9][0-9_]*)\\.[0-9_]*(?:[eE][-+]?[0-9]+)?
            |[-+]?(?:[0-9][0-9_]*)(?:[eE][-+]?[0-9]+)
            |\\.[0-9_]+(?:[eE][-+][0-9]+)?
            |[-+]?[0-9][0-9_]*(?::[0-5]?[0-9])+\\.[0-9_]*
            |[-+]?\\.(?:inf|Inf|INF)
            |\\.(?:nan|NaN|NAN))$''', re.X
            ), list(u'-+0123456789.')
        )
        return loader

    def dot_contains(self, dot_string: str):
        keys = [self.__config__] + dot_string.strip().split('.')
        try:
            reduce(lambda x, y: x[y], keys)
        except Exception:
            return False
        return True

    def dot_get(self, dot_string: str, default_value: Any = None, require=False):
        keys = [self.__config__] + dot_string.strip().split('.')
        if not require:
            try:
                return reduce(lambda x, y: x[y], keys)
            except KeyError:
                return default_value
        else:
            return reduce(lambda x, y: x[y], keys)

    def dot_set(self, dot_string: str, value: Any = None):
        keys = dot_string.strip().split('.')
        obj = reduce(lambda x, y: x[y], [self.__config__] + keys[:-1])
        obj[keys[-1]] = value

    def __iter__(self):
        for k in self.__config__.keys():
            yield k

    def __getattr__(self, key: str):
        if '__config__' in self.__dict__ and key in self.__dict__['__config__']:
            return self.__dict__['__config__'][key]
        elif key in dir(self):
            return self.__dict__[key]
        else:
            raise AttributeError(f"'{self.__class__.__name__}' object has no attribute '{key}'")

    def __setattr__(self, key: str, value: Any):
        if key != "__config__":
            self.__config__[key] = value
        else:
            self.__dict__[key] = value

    def __delattr__(self, key: str):
        if key in self.__config__.keys():
            del self.__config__[key]
        elif key in dir(self):
            raise AttributeError(f"attribute '{key}' is not allowed to delete")
        else:
            raise AttributeError(f"'{self.__class__.__base__}' object has no attribute '{key}'")

    def __setitem__(self, key: str, value: Any):
        assert key not in dir(self), "conflict with dir(self)"
        self.__config__[key] = value

    def __getitem__(self, key: str):
        return self.__config__[key]

    def __delitem__(self, key: str):
        del self.__config__[key]

    def keys(self):
        return self.__config__.keys()

    def items(self):
        return self.__config__.items()

    def to_dict(self):
        return copy.deepcopy(self.__config__)

    def get(self, key: str, default_value: Any = None):
        return self.__config__.get(key, default_value)

    def update(self, dict_obj: Union[dict, object], update_unknown_key_only=False):
        for k in dict_obj:
            if k in self and update_unknown_key_only:
                continue
            self[k] = dict_obj[k]

    def __str__(self):
        return f"{self.__class__.__name__}({self.__config__})"

    def __repr__(self):
        return self.__str__()

    def dump_fmt(self):
        return json.dumps(
            self.__config__, indent=4, ensure_ascii=False,
            cls=NumpyEncoder
        )

    def dump_file(self, filepath: str, encoding: str = 'utf-8'):
        with open(filepath, "w", encoding=encoding) as f:
            json.dump(
                self.__config__, f, indent=4, ensure_ascii=False,
                cls=NumpyEncoder
            )

    def __copy__(self):
        cls = self.__class__
        return cls(dic=copy.copy(self.__config__))

    def __deepcopy__(self, memo: Any):
        cls = self.__class__
        return cls(dic=copy.deepcopy(self.__config__, memo=memo))


class NumpyEncoder(json.JSONEncoder):
    """ 
    Custom encoder for numpy data types 
    Ref: https://github.com/hmallen/numpyencoder/blob/f8199a61ccde25f829444a9df4b21bcb2d1de8f2/numpyencoder/numpyencoder.py
    """

    def default(self, obj):
        if isinstance(obj, UnifyConfig):
            return obj.__config__
        elif isinstance(obj, (np.int_, np.intc, np.intp, np.int8,
                            np.int16, np.int32, np.int64, np.uint8,
                            np.uint16, np.uint32, np.uint64)):
            return int(obj)

        elif isinstance(obj, (np.float16, np.float32, np.float64)):
            return float(obj)

        elif isinstance(obj, (np.complex64, np.complex128)):
            return {'real': obj.real, 'imag': obj.imag}

        elif isinstance(obj, (np.ndarray,)):
            return obj.tolist()

        elif isinstance(obj, (np.bool_)):
            return bool(obj)

        elif isinstance(obj, (np.void)): 
            return None
        try:
            return json.JSONEncoder.default(self, obj)
        except TypeError:
            return str(obj)



from ..gd_basemodel import GDBaseModel
import torch.nn as nn
import torch
import torch.nn.functional as F

class DKVMNHeadGroup(nn.Module):
    def forward(self, input_):
        pass

    def __init__(self, memory_size, memory_state_dim, is_write, device):
        super(DKVMNHeadGroup, self).__init__()
        """"
        Parameters
            memory_size:        scalar
            memory_state_dim:   scalar
            is_write:           boolean
        """
        self.memory_size = memory_size
        self.memory_state_dim = memory_state_dim
        self.is_write = is_write
        if self.is_write:
            self.erase = nn.Linear(self.memory_state_dim, self.memory_state_dim, bias=True)
            self.add = nn.Linear(self.memory_state_dim, self.memory_state_dim, bias=True)
            nn.init.kaiming_normal_(self.erase.weight)
            nn.init.kaiming_normal_(self.add.weight)
            nn.init.constant_(self.erase.bias, 0)
            nn.init.constant_(self.add.bias, 0)
        self.device = device

    @staticmethod
    def addressing(control_input, memory):
        """
        Parameters
            control_input:          Shape (batch_size, control_state_dim)
            memory:                 Shape (memory_size, memory_state_dim)
        Returns
            correlation_weight:     Shape (batch_size, memory_size)
        """
        similarity_score = torch.matmul(control_input, torch.t(memory))
        correlation_weight = F.softmax(similarity_score, dim=1)  # Shape: (batch_size, memory_size)
        
        return correlation_weight
    
    def read(self, memory, control_input=None, read_weight=None):
        """
        Parameters
            control_input:  Shape (batch_size, control_state_dim)
            memory:         Shape (batch_size, memory_size, memory_state_dim)
            read_weight:    Shape (batch_size, memory_size)
        Returns
            read_content:   Shape (batch_size,  memory_state_dim)
        """
        if read_weight is None:
            read_weight = self.addressing(control_input=control_input, memory=memory)
        read_weight = read_weight.view(-1, 1) 
        memory = memory.view(-1, self.memory_state_dim)
        rc = torch.mul(read_weight, memory)  
        read_content = rc.view(-1, self.memory_size, self.memory_state_dim)
        read_content = torch.sum(read_content, dim=1)

        return read_content
    def write(self, control_input, memory, write_weight=None):
        """
        Parameters
            control_input:      Shape (batch_size, control_state_dim)
            write_weight:       Shape (batch_size, memory_size)
            memory:             Shape (batch_size, memory_size, memory_state_dim)
        Returns
            new_memory:         Shape (batch_size, memory_size, memory_state_dim)
        """
        assert self.is_write
        if write_weight is None:
            write_weight = self.addressing(control_input=control_input, memory=memory)
        erase_signal = torch.sigmoid(self.erase(control_input))
        add_signal = torch.tanh(self.add(control_input))
        erase_reshape = erase_signal.view(-1, 1, self.memory_state_dim)
        add_reshape = add_signal.view(-1, 1, self.memory_state_dim)
        write_weight_reshape = write_weight.view(-1, self.memory_size, 1)
        erase_mul = torch.mul(erase_reshape, write_weight_reshape)
        add_mul = torch.mul(add_reshape, write_weight_reshape)
        new_memory = memory * (1 - erase_mul) + add_mul

        return new_memory

class DKVMN(nn.Module):
    def forward(self, input_):
        pass

    def __init__(self, memory_size, memory_key_state_dim, memory_value_state_dim, init_memory_key, device):
        super(DKVMN, self).__init__()
        """
        :param memory_size:             scalar
        :param memory_key_state_dim:    scalar
        :param memory_value_state_dim:  scalar
        :param init_memory_key:         Shape (memory_size, memory_value_state_dim)
        """
        self.memory_size = memory_size
        self.memory_key_state_dim = memory_key_state_dim
        self.memory_value_state_dim = memory_value_state_dim
        self.key_head = DKVMNHeadGroup(memory_size=self.memory_size,
                                       memory_state_dim=self.memory_key_state_dim,
                                       is_write=False,
                                       device=device)

        self.value_head = DKVMNHeadGroup(memory_size=self.memory_size,
                                         memory_state_dim=self.memory_value_state_dim,
                                         is_write=True,
                                         device=device)
        self.memory_key = init_memory_key
        self.device = device

    def attention(self, control_input):
        correlation_weight = self.key_head.addressing(control_input=control_input, memory=self.memory_key)
        
        return correlation_weight

    def read(self, read_weight, memory_value):
        read_content = self.value_head.read(memory=memory_value, read_weight=read_weight)

        return read_content

    def write(self, write_weight, control_input, memory_value):
        memory_value = self.value_head.write(control_input=control_input,
                                             memory=memory_value,
                                             write_weight=write_weight,
                                             )
        
        return memory_value

class SKVMN(GDBaseModel):
    default_cfg = {
        'memory_size': 50,
        'embed_dim': 200,  
        'dropout': 0.2,
    }

    def __init__(self, cfg):
        super().__init__(cfg)

    def build_cfg(self):
        self.n_user = self.datatpl_cfg['dt_info']['stu_count']
        self.n_item = self.datatpl_cfg['dt_info']['exer_count']
        self.window_size = self.datatpl_cfg['dt_info']['real_window_size']
        

    def build_model(self):
        self.x_emb_layer = nn.Embedding(2 * self.n_item, self.modeltpl_cfg['embed_dim'])
        self.k_emb_layer = nn.Embedding(self.n_item, self.modeltpl_cfg['embed_dim'])
        self.Mk = nn.Parameter(torch.Tensor(self.modeltpl_cfg['memory_size'], self.modeltpl_cfg['embed_dim']))
        self.Mv = nn.Parameter(torch.Tensor(self.modeltpl_cfg['memory_size'], self.modeltpl_cfg['embed_dim']))

        nn.init.kaiming_normal_(self.Mk)
        nn.init.kaiming_normal_(self.Mv)

        self.read_embed_linear = nn.Linear(
            2 * self.modeltpl_cfg['embed_dim'],
            self.modeltpl_cfg['embed_dim'],
            bias=True
        )
        self.a_embed = nn.Linear(
            2 * self.modeltpl_cfg['embed_dim'],
            self.modeltpl_cfg['embed_dim'],
            bias=True
            )
        self.mem = DKVMN(
            memory_size=self.modeltpl_cfg['memory_size'],
            memory_key_state_dim=self.modeltpl_cfg['embed_dim'],
            memory_value_state_dim=self.modeltpl_cfg['embed_dim'],
            init_memory_key=self.Mk,
            device=self.traintpl_cfg['device']
        )
        self.hx = nn.Parameter(torch.Tensor(1, self.modeltpl_cfg['embed_dim']))
        self.cx = nn.Parameter(torch.Tensor(1, self.modeltpl_cfg['embed_dim']))
        nn.init.kaiming_normal_(self.hx )
        nn.init.kaiming_normal_(self.cx)
        # modify hop_lstm
        self.lstm_cell = nn.LSTMCell(self.modeltpl_cfg['embed_dim'], self.modeltpl_cfg['embed_dim'])
        self.dropout_layer = nn.Dropout(self.modeltpl_cfg['dropout'])
        self.p_layer = nn.Linear(self.modeltpl_cfg['embed_dim'], 1, bias=True)

    def ut_mask(self, seq_len):
        return torch.triu(torch.ones(seq_len, seq_len), diagonal=0).to(dtype=torch.bool)
    
    def triangular_layer(self, correlation_weight, seqlen=100, a=0.075, b=0.088, c=1.00):
        # w'= min((w-a)/(b-a), (c-w)/(c-b))
        # man(w', 0)
        correlation_weight = correlation_weight.view(self.bs * seqlen, -1)
        correlation_weight = torch.cat([correlation_weight[i] for i in range(correlation_weight.shape[0])], 0).unsqueeze(0)
        correlation_weight = torch.cat([(correlation_weight-a)/(b-a), (c-correlation_weight)/(c-b)], 0)
        correlation_weight, _ = torch.min(correlation_weight, 0)
        w0 = torch.zeros(correlation_weight.shape[0]).to(self.traintpl_cfg['device'])
        correlation_weight = torch.cat([correlation_weight.unsqueeze(0), w0.unsqueeze(0)], 0)
        correlation_weight, _ = torch.max(correlation_weight, 0)

        identity_vector_batch = torch.zeros(correlation_weight.shape[0]).to(self.traintpl_cfg['device'])
        # >=0.6 set to 2，0.1-0.6 set to 1，<=0.1 set to 0
        identity_vector_batch = identity_vector_batch.masked_fill(correlation_weight.lt(0.1), 0)
        identity_vector_batch = identity_vector_batch.masked_fill(correlation_weight.ge(0.1), 1)
        _identity_vector_batch = identity_vector_batch.masked_fill(correlation_weight.ge(0.6), 2)

        identity_vector_batch = _identity_vector_batch.view(self.bs * seqlen, -1)
        identity_vector_batch = torch.reshape(identity_vector_batch,[self.bs, seqlen, -1])  #u(x): [bs, seqlen, size_m]
        # A^2
        iv_square_norm = torch.sum(torch.pow(identity_vector_batch, 2), dim=2, keepdim=True)
        iv_square_norm = iv_square_norm.repeat((1, 1, iv_square_norm.shape[1]))
        # B^2.T
        unique_iv_square_norm = torch.sum(torch.pow(identity_vector_batch, 2), dim=2, keepdim=True)
        unique_iv_square_norm = unique_iv_square_norm.repeat((1, 1, seqlen)).transpose(2, 1)
        # A * B.T
        iv_matrix_product = torch.bmm(identity_vector_batch, identity_vector_batch.transpose(2,1)) # A * A.T 
        # A^2 + B^2 - 2A*B.T
        iv_distances = iv_square_norm + unique_iv_square_norm - 2 * iv_matrix_product
        iv_distances = torch.where(iv_distances>0.0, torch.tensor(-1e32).to(self.traintpl_cfg['device']), iv_distances) 
        masks = self.ut_mask(iv_distances.shape[1]).to(self.traintpl_cfg['device'])
        mask_iv_distances = iv_distances.masked_fill(masks, value=torch.tensor(-1e32).to(self.traintpl_cfg['device'])) 
        idx_matrix = torch.arange(0,seqlen * seqlen,1).reshape(seqlen,-1).repeat(self.bs,1,1).to(self.traintpl_cfg['device'])
        final_iv_distance = mask_iv_distances + idx_matrix 
        values, indices = torch.topk(final_iv_distance, 1, dim=2, largest=True) 

        _values = values.permute(1,0,2)
        _indices = indices.permute(1,0,2)
        batch_identity_indices = (_values >= 0).nonzero() 
        identity_idx = []
        for identity_indices in batch_identity_indices:
            pre_idx = _indices[identity_indices[0],identity_indices[1]] 
            idx = torch.cat([identity_indices[:-1],pre_idx], dim=-1)
            identity_idx.append(idx)
        if len(identity_idx) > 0:
            identity_idx = torch.stack(identity_idx, dim=0)
        else:
            identity_idx = torch.tensor([]).to(self.traintpl_cfg['device'])

        return identity_idx 

    def forward(self, exer_seq, label_seq, **kwargs):
        k = self.k_emb_layer(exer_seq)  # Embedding Layer: about question
        self.bs = exer_seq.shape[0]
        x = (exer_seq + self.n_item * label_seq).long()

        value_read_content_l = []  # seqlen * bs * dim | read_content
        input_embed_l = []  # seqlen * bs * dim| q
        correlation_weight_list = []  # seqlen * bs * mem_size | correlation_weight
        ft = []
        # After atten about new question, Update Memory-value
        mem_value = self.Mv.unsqueeze(0).repeat(self.bs, 1, 1).to(self.traintpl_cfg['device'])  # bs * mem_size * dim
        for i in range(self.window_size):
            # k: bs * seqlen * dim
            q = k.permute(1, 0, 2)[i]  # q: bs * dim
            # attention Process
            correlation_weight = self.mem.attention(q).to(self.traintpl_cfg['device'])  # bs * memory_size
            
            # Read Process
            read_content = self.mem.read(correlation_weight, mem_value)  # read_count: bs * dim
            # modify
            batch_predict_input = torch.cat([read_content, q], 1)  #  bs * 2dim
            f = torch.tanh(self.read_embed_linear(batch_predict_input))

            # save intermedium data
            correlation_weight_list.append(correlation_weight)
            value_read_content_l.append(read_content)
            input_embed_l.append(q)
            ft.append(f)

            # Write Process  |  student response + f-vector 
            y = self.x_emb_layer(x[:,i])
            write_embed = torch.cat([f, y], 1) # bz * (f_dim + y_dim)
            write_embed = self.a_embed(write_embed).to(self.traintpl_cfg['device']) # bs * dim
            mem_value = self.mem.write(correlation_weight, write_embed, mem_value)
        # w: bs * seqlen * mem_size
        w = torch.cat([correlation_weight_list[i].unsqueeze(1) for i in range(self.window_size)], 1)
        idx_values = self.triangular_layer(w, self.window_size)
        ft = torch.stack(ft, dim=0)

        hidden_state, cell_state = [], []
        hx, cx = self.hx.repeat(self.bs, 1), self.cx.repeat(self.bs, 1)
        for i in range(self.window_size):
            for j in range(self.bs):
                if idx_values.shape[0] != 0 and i == idx_values[0][0] and j == idx_values[0][1]:
                    hx[j,:] = hidden_state[idx_values[0][2]][j]
                    cx = cx.clone()
                    cx[j,:] = cell_state[idx_values[0][2]][j]
                    idx_values = idx_values[1:]
            hx, cx = self.lstm_cell(ft[i], (hx, cx)) 
            hidden_state.append(hx) 
            cell_state.append(cx) 
        hidden_state = torch.stack(hidden_state, dim=0).permute(1,0,2)
        cell_state = torch.stack(cell_state, dim=0).permute(1,0,2)

        p = self.p_layer(self.dropout_layer(hidden_state))
        p = torch.sigmoid(p)
        y_pred = p.squeeze(-1)

        return y_pred
    
    @torch.no_grad()
    def predict(self, **kwargs):
        y_pd = self(**kwargs)
        y_pd = y_pd[:, :-1]
        y_pd = y_pd[kwargs['mask_seq'][:, 1:] == 1]
        y_gt = None
        if kwargs.get('label_seq', None) is not None:
            y_gt = kwargs['label_seq'][:, 1:]
            y_gt = y_gt[kwargs['mask_seq'][:, 1:] == 1]
        return {
            'y_pd': y_pd,
            'y_gt': y_gt
        }

    def get_main_loss(self, **kwargs):
        y_pd = self(**kwargs)
        y_pd = y_pd[:, :-1]
        y_pd = y_pd[kwargs['mask_seq'][:, 1:] == 1]
        y_gt = kwargs['label_seq'][:, 1:]
        y_gt = y_gt[kwargs['mask_seq'][:, 1:] == 1]
        loss = F.binary_cross_entropy(
            input=y_pd, target=y_gt
        )
        return {
            'loss_main': loss
        }

    def get_loss_dict(self, **kwargs):
        return self.get_main_loss(**kwargs)



from .raw2mid import BaseRaw2Mid
import pandas as pd
r"""
R2M_NIPS12
#####################################
NIPS12 dataset preprocess
"""


class R2M_Eedi_20_T12(BaseRaw2Mid):
    """R2M_Eedi_20_T12 is to preprocess NIPS 2020 challenge Task 1&2 dataset"""
    def process(self):
        super().process()
        # 读入数据 查看
        df = pd.read_csv(f"{self.rawpath}/train_data/train_task_1_2.csv",encoding='utf-8')
        df_exer = pd.read_csv(f"{self.rawpath}/metadata/answer_metadata_task_1_2.csv", encoding='utf-8')
        df_qus = pd.read_csv(f"{self.rawpath}/metadata/question_metadata_task_1_2.csv", encoding='utf-8')
        df_user = pd.read_csv(f"{self.rawpath}/metadata/student_metadata_task_1_2.csv", encoding='utf-8')
        df_rel = pd.read_csv(f"{self.rawpath}/metadata/subject_metadata.csv", encoding='utf-8')

        # 合并，获取 DateAnswered 等信息
        df = pd.merge(df, df_exer[['AnswerId', 'DateAnswered', 'GroupId', 'SchemeOfWorkId']], on='AnswerId', how='left')
        df = pd.merge(df,df_user[['UserId', 'Gender']],on='UserId', how='left')
        # # print(df.head())
        df = pd.merge(df,df_qus[['QuestionId','SubjectId']], on='QuestionId',how='left')
        df.drop(['AnswerId', 'CorrectAnswer', 'AnswerValue'], axis=1, inplace=True)
        # print(df.head())

        # 处理学生多个班级的问题，保留做题数目最多的班级
        class_problem_counts = df.groupby(['GroupId', 'UserId'])['QuestionId'].nunique().reset_index()
        max_class_counts = class_problem_counts.groupby('UserId')['QuestionId'].idxmax()
        max_class_counts = class_problem_counts.loc[max_class_counts, ['UserId', 'GroupId']]
        df = df.merge(max_class_counts, on=['UserId', 'GroupId'])

        # 处理性别 不确定为空 男1 女0
        mapping = {0:None, 1:0, 2:1}
        df['Gender'] = df['Gender'].map(mapping)

        # 处理答题时间
        df['DateAnswered'] = pd.to_datetime(df['DateAnswered'])
        base_time = df['DateAnswered'].min()
        df['timestamp'] = (df['DateAnswered'] - base_time).dt.total_seconds() * 1000
        # print(df)

        # 进行映射
        def sort(data, column):
            '''将原始数据对指定列进行排序，并完成0-num-1映射'''
            if(column != 'DateAnswered'):
                sorted_id = sorted(data[column].unique())
                num = len(sorted_id)
                mapping = {id: i for i, id in enumerate(sorted_id)}
                data[column] = data[column].replace('', '').map(mapping)
                
            # 单独处理 DateAnswered，用字典映射
            else: 
                data = data.sort_values(by=['UserId', 'DateAnswered'])

                user_mapping = {}
                user_count = {}

                def generate_mapping(row):
                    user_id = row['UserId']
                    timestamp = row['DateAnswered']
                    if user_id not in user_mapping:
                        user_mapping[user_id] = {}
                        user_count[user_id] = 0
                    if timestamp not in user_mapping[user_id]:
                        user_mapping[user_id][timestamp] = user_count[user_id]
                        user_count[user_id] += 1
                    return user_mapping[user_id][timestamp]
                data['new_order_id'] = data.apply(generate_mapping, axis=1)
            return data

        df = sort(df,'UserId')
        df = sort(df,'QuestionId')
        df = sort(df,'timestamp')
        df = sort(df,'GroupId')
        df = sort(df,'SchemeOfWorkId')
        df = sort(df,'DateAnswered')

        # 映射cpt中的内容
        sorted_cpt_id = sorted(df_rel['SubjectId'].unique())
        mapping = {cpt_id: i for i, cpt_id in enumerate(sorted_cpt_id)}
        df_rel['SubjectId'] = df_rel['SubjectId'].map(mapping)
        df_rel['ParentId'] = df_rel['ParentId'].replace('', '').map(mapping)

        # 修改列名及顺序
        df = df.rename(columns = {'UserId' : 'stu_id:token', 'SchemeOfWorkId':'assignment_id:token_seq',
                                  'QuestionId' : 'exer_id:token', 'IsCorrect' : 'label:float','GroupId':'class_id:token',
                                  'timestamp':'start_timestamp:float','Gender':'gender:float','SubjectId':'cpt_seq:token_seq',
                                 'new_order_id':'order_id:float'})
        new_column_order = ['stu_id:token','exer_id:token','label:float','start_timestamp:float','order_id:float',
                            'class_id:token', 'gender:float','cpt_seq:token_seq','assignment_id:token_seq']
        df = df.reindex(columns=new_column_order)

        # df_inter 的相关处理
        df_inter = df[['stu_id:token','exer_id:token','label:float','start_timestamp:float','order_id:float']]
        df_inter.drop_duplicates(inplace=True)
        df_inter .sort_values(by = ['stu_id:token','order_id:float'], inplace=True)

        # df_user 相关处理
        df_user = df[['stu_id:token','class_id:token','gender:float']]
        df_user.drop_duplicates(inplace=True)
        df_user .sort_values('stu_id:token', inplace=True)

        # df_exer 相关处理
        df_exer = df[['exer_id:token','cpt_seq:token_seq','assignment_id:token_seq']]
        df_exer.drop_duplicates(inplace=True)
        # 拆分知识点列
        df_exer['cpt_seq:token_seq'] = df_exer['cpt_seq:token_seq'].str.strip('[]') 
        df_exer['cpt_seq:token_seq'] = df_exer['cpt_seq:token_seq'].str.split(', ')  
        df_exer = df_exer.explode('cpt_seq:token_seq') 
        df_exer = sort(df_exer,'cpt_seq:token_seq')
        # 合并 cpt_seq
        grouped_skills = df_exer[['exer_id:token','cpt_seq:token_seq']]
        grouped_skills.drop_duplicates(inplace=True)
        grouped_skills.sort_values(by='cpt_seq:token_seq', inplace=True)
        grouped_skills['exer_id:token'] = grouped_skills['exer_id:token'].astype(str)
        grouped_skills['cpt_seq:token_seq'] = grouped_skills['cpt_seq:token_seq'].astype(str)
        grouped_skills  = grouped_skills.groupby('exer_id:token')['cpt_seq:token_seq'].agg(','.join).reset_index()
        # 合并 assignment_id
        grouped_assignments = df_exer[['exer_id:token','assignment_id:token_seq']]
        grouped_assignments.drop_duplicates(inplace=True)
        grouped_assignments.sort_values(by='assignment_id:token_seq', inplace=True)
        grouped_assignments['exer_id:token'] = grouped_assignments['exer_id:token'].astype(str)
        grouped_assignments['assignment_id:token_seq'] = grouped_assignments['assignment_id:token_seq'].astype(str)
        grouped_assignments  = grouped_assignments.groupby('exer_id:token')['assignment_id:token_seq'].agg(','.join).reset_index()
        # 结果合并
        df_exer = pd.merge(grouped_skills, grouped_assignments, on='exer_id:token', how='left')
        df_exer['exer_id:token'] = df_exer['exer_id:token'].astype(int)
        df_exer.sort_values(by='exer_id:token', inplace=True)

        # df_cpt 相关处理
        # 读取信息完成映射
        df_cpt = df_rel[['SubjectId','ParentId','Level']]
        df_cpt = df_cpt.rename(columns={'SubjectId':'cpt_id:token', 'ParentId':'parent_id:token','Level':'level:float'})
        df_cpt['cpt_id:token'] = df_cpt['cpt_id:token'].astype(int)
        df_cpt.sort_values(by='cpt_id:token',inplace=True)

        # 保存数据
        df_inter.to_csv(f"{self.midpath}/{self.dt}.inter.csv", index=False, encoding='utf-8')
        df_user.to_csv(f"{self.midpath}/{self.dt}.user.csv", index=False, encoding='utf-8')
        df_exer.to_csv(f"{self.midpath}/{self.dt}.exer.csv", index=False, encoding='utf-8')
        df_cpt.to_csv(f"{self.midpath}/{self.dt}.cpt_tree.csv", index=False, encoding='utf-8')


import os
import re
import glob
from .raw2mid import BaseRaw2Mid
import pandas as pd
r"""
R2M_Simulated5
#####################################
Simulated5 dataset preprocess
"""


class R2M_Simulated5(BaseRaw2Mid):
    """R2M_Simulated5 is to preprocess Simulated5 dataset"""
    def process(self):
        super().process()

        def process_inter(path):
            # 读取数据，获得label
            df_inter = pd.read_csv(path, encoding='utf-8').reset_index()
            del df_inter['50']
            df_inter.columns = range(len(df_inter.columns))

            # 插入一行使其成为3的整数倍
            df_inter.index = df_inter.index + 1
            df_inter.loc[0] = 50
            df_inter = df_inter.reset_index(drop=True)

            # 对数据进行切片
            df_inter = df_inter.iloc[1::3]
            df_inter = df_inter.reset_index(drop=True)
            df_inter.insert(0, 'stu_id:token', range(len(df_inter)))

            # 拆成交互数据
            df_inter = df_inter.melt(id_vars=['stu_id:token'], var_name='exer_id:token', value_name='label:float')
            df_inter.sort_values(by=['stu_id:token', 'exer_id:token'], inplace=True)
            return df_inter

        def process_exer(path):
            # 读取文本文件，提取所需信息
            path = f'{self.rawpath}/cluster_info_0'
            with open(path, 'r') as file:
                content = file.read().replace('\n', '').replace('\t', '')
            pattern = r'\[(.*?)\]'
            matches = re.findall(pattern, content)
            results = [match.split(', ') for match in matches][1:]
            df_exer = pd.DataFrame([(exer, cpt) for cpt, exers in enumerate(results) for exer in exers],
                                   columns=['exer_id:token', 'cpt_id:token'])
            df_exer['exer_id:token'] = df_exer['exer_id:token'].astype(int) - 1
            df_exer = df_exer.sort_values(by='exer_id:token')
            return df_exer

        # 处理后数据保存的文件夹路径
        output_folder = f"{self.midpath}"

        # 文件名模式
        file_pattern_inter = f"{self.rawpath}/naive_c5_q50_s4000_v*_train.csv"
        file_path_inters = glob.glob(file_pattern_inter)

        for file_path_inter in file_path_inters:
            processed_data = process_inter(file_path_inter)
            # 获取文件名
            file_name = f'{self.dt}_' + os.path.basename(file_path_inter)
            # 构建处理后数据的保存路径
            output_path = os.path.join(output_folder, file_name)
            # 生成新的文件名
            new_filename = output_path.replace(".csv", ".inter.csv")
            processed_data.to_csv(new_filename, index=False, encoding='utf-8')

        # 文件名模式
        file_pattern_exer = f"{self.rawpath}/cluster_info_*"
        file_path_exers = glob.glob(file_pattern_exer)

        for file_path_exer in file_path_exers:
            processed_data = process_exer(file_path_exer)
            file_name = f'{self.dt}_' + os.path.basename(file_path_exer)
            output_path = os.path.join(output_folder, file_name)
            new_file_name = output_path + '.exer.csv'
            processed_data.to_csv(new_file_name, index=False, encoding='utf-8')



r"""
IRR
##################################
Reference:
    Tong et al. "Item Response Ranking for Cognitive Diagnosis." in IJCAI 2021.
"""
from ..basemodel import BaseProxyModel
import torch
import torch.nn as nn
import torch.nn.functional as F


class PairSCELoss(nn.Module):
    """IRR loss function"""
    def __init__(self):
        """IRR loss will use cross entropy"""
        super(PairSCELoss, self).__init__()
        self._loss = nn.CrossEntropyLoss()

    def forward(self, pred1, pred2, sign=1, *args):
        """Get the PairSCELoss

        Args:
            pred1 (torch.Tensor): positive prediction
            pred2 (_type_): negtive prediction
            sign (int, optional): 1: pred1 should be greater than pred2; -1: otherwise. Defaults to 1.

        Returns:
            torch.Tensor: PairSCELoss
        """
        pred = torch.stack([pred1, pred2], dim=1)
        return self._loss(pred, ((torch.ones(pred1.shape[0], device=pred.device) - sign) / 2).long())



class IRR(BaseProxyModel):
    """
    backbone_modeltpl_cls: The backbone model of IRR
    """
    default_cfg = {
        "backbone_modeltpl_cls": "IRT",
        'pair_weight': 0.5,
    }

    def __init__(self, cfg):
        """Pass parameters from other templates into the model

        Args:
            cfg (UnifyConfig): parameters from other templates
        """
        super().__init__(cfg)

    def build_model(self):
        """Initialize the various components of the model"""
        super().build_model()
        self.irr_pair_loss = PairSCELoss()

    def get_main_loss(self, **kwargs):
        """Get the loss of IRR

        Returns:
            dict: {'loss_main': loss}
        """
        pair_exer = kwargs['pair_exer']
        pair_pos_stu = kwargs['pair_pos_stu']
        pair_neg_stu = kwargs['pair_neg_stu']
        
        kwargs['exer_id'] = pair_exer
        kwargs['stu_id'] = pair_pos_stu
        pos_pd = self(**kwargs).flatten()
        kwargs['stu_id'] = pair_neg_stu
        neg_pd = self(**kwargs).flatten()
        pos_label = torch.ones(pos_pd.shape[0]).to(self.device)
        neg_label = torch.zeros(neg_pd.shape[0]).to(self.device)
        point_loss = F.binary_cross_entropy(input=pos_pd, target=pos_label) + F.binary_cross_entropy(input=neg_pd, target=neg_label)

        return {
            'loss_main': self.modeltpl_cfg['pair_weight'] * self.irr_pair_loss(pos_pd, neg_pd) + (1-self.modeltpl_cfg['pair_weight']) * point_loss
        }



from edustudio.utils.common import UnifyConfig
import logging


class BaseMid2Cache(object):
    default_cfg = {}

    def __init__(self, m2c_cfg) -> None:
        self.logger = logging.getLogger("edustudio")
        self.m2c_cfg = m2c_cfg
        self._check_params()

    def _check_params(self):
        pass

    @classmethod
    def from_cfg(cls, cfg: UnifyConfig):
        return cls(cfg.datatpl_cfg.get(cls.__name__))
    
    @classmethod
    def get_default_cfg(cls, **kwargs):
        cfg = UnifyConfig()
        for _cls in cls.__mro__:
            if not hasattr(_cls, 'default_cfg'):
                break
            cfg.update(_cls.default_cfg, update_unknown_key_only=True)
        return cfg
    
    def process(self, **kwargs):
        pass

    def set_dt_info(self, dt_info, **kwargs):
        pass



# -*- coding: utf-8 -*-
# @Author : Xiangzhi Chen
# @Github : kervias

from .callback import Callback
from collections import defaultdict


class BaseLogger(Callback):
    def __init__(self, logger=None, **kwargs):
        super(BaseLogger, self).__init__()
        self.log = print
        if logger is not None and hasattr(logger, 'info'):
            self.log = logger.info

        self.join_str = kwargs.get("join_str", " | ")
        self.group_by_contains = kwargs.get('group_by_contains', ('loss'))
        self.group_by_count = kwargs.get('group_by_count', 5)
        assert self.group_by_count >= 1 and type(self.group_by_count) is int

    def on_train_begin(self, logs=None, **kwargs):
        super().on_train_begin()
        self.log("Start Training...")

    def on_epoch_end(self, epoch: int, logs: dict = None, **kwargs):
        info = f"[EPOCH={epoch:03d}]: "
        flag = [False] * len(logs)
        for group_rule in self.group_by_contains:
            v_list = []
            for i, (k, v) in enumerate(logs.items()):
                if group_rule in k and not flag[i]:
                    v_list.append(f"{k}: {v:.4f}")
                    flag[i] = True
            if len(v_list) > 0: self.log(f"{info}{self.join_str.join(v_list)}")
        
        logs = {k: logs[k] for i, k in enumerate(logs) if not flag[i]}

        v_list = []
        for i, (k, v) in enumerate(logs.items()):
            v_list.append(f"{k}: {v:.4f}")
            if (i+1) % self.group_by_count == 0:
                self.log(f"{info}{self.join_str.join(v_list)}")
                v_list = []
        if len(v_list) > 0: self.log(f"{info}{self.join_str.join(v_list)}")

    def on_train_end(self, logs=None, **kwargs):
        self.log("Training Completed!")



from .raw2mid import BaseRaw2Mid
import pandas as pd
r"""
R2M_Math1
#####################################
Math1 dataset preprocess
"""


class R2M_Math1(BaseRaw2Mid):
    """R2M_Math1 is to preprocess Math1 dataset"""
    def process(self):
        super().process()
        # 读取文本文件转换为 dataframe
        df_inter = pd.read_csv(f"{self.rawpath}/data.txt", sep='\t', names=['0', '1', '2', '3', '4', '5', '6', '7', '8', '9',
                                                                        '10', '11', '12', '13', '14', '15', '16', '17', '18', '19']) 
        df_exer = pd.read_csv(f"{self.rawpath}/q.txt", sep='\t', names=['0', '1', '2', '3', '4', '5', '6', '7', '8', '9','10'])

        # 处理 df_inter 拆成很多行
        df_inter.insert(0, 'stu_id:token', range(len(df_inter)))
        df_inter = df_inter.melt(id_vars=['stu_id:token'], var_name='exer_id:token', value_name='label:float')
        df_inter['exer_id:token'] = df_inter['exer_id:token'].astype(int)
        df_inter .sort_values(by = ['stu_id:token','exer_id:token'],inplace=True)

        # 处理 df_exer 先拆成很多行，再合并
        
        # 拆成很多行
        df_exer.insert(0, 'exer_id:token', range(len(df_exer)))
        df_exer = df_exer.melt(id_vars=['exer_id:token'], var_name='cpt_seq:token_seq', value_name='value')
        df_exer = df_exer[df_exer['value'] == 1]
        del df_exer['value']

        # 合并 cpt_seq:token_seq
        df_exer['cpt_seq:token_seq'] = df_exer['cpt_seq:token_seq'].astype(int)
        df_exer.sort_values(by='cpt_seq:token_seq', inplace=True)
        df_exer['exer_id:token'] = df_exer['exer_id:token'].astype(str)
        df_exer['cpt_seq:token_seq'] = df_exer['cpt_seq:token_seq'].astype(str)
        df_exer  = df_exer.groupby('exer_id:token')['cpt_seq:token_seq'].agg(','.join).reset_index()

        # 按 exer_id:token 进行排序
        df_exer['exer_id:token'] = df_exer['exer_id:token'].astype(int)
        df_exer.sort_values(by='exer_id:token', inplace=True)

        # 此处将数据保存到`self.midpath`中

        df_inter.to_csv(f"{self.midpath}/{self.dt}.inter.csv", index=False, encoding='utf-8')
        df_exer.to_csv(f"{self.midpath}/{self.dt}.exer.csv", index=False, encoding='utf-8')



# -*- coding: utf-8 -*-
# @Author : Xiangzhi Chen
# @Github : kervias

import logging
from edustudio.utils.common import UnifyConfig
from edustudio.utils.callback import CallbackList


class BaseEvalTPL(object):
    """The baisc protocol for implementing a evaluate template
    """
    default_cfg = {}

    def __init__(self, cfg):
        self.cfg: UnifyConfig = cfg
        self.datatpl_cfg: UnifyConfig = cfg.datatpl_cfg
        self.evaltpl_cfg: UnifyConfig = cfg.evaltpl_cfg
        self.traintpl_cfg: UnifyConfig = cfg.traintpl_cfg
        self.frame_cfg: UnifyConfig = cfg.frame_cfg
        self.modeltpl_cfg: UnifyConfig = cfg.modeltpl_cfg
        self.logger: logging.Logger = logging.getLogger("edustudio")
        self.name = self.__class__.__name__
        self._check_params()

    @classmethod
    def get_default_cfg(cls):
        parent_class = cls.__base__
        cfg = UnifyConfig(cls.default_cfg)
        if hasattr(parent_class, 'get_default_cfg'):
            cfg.update(parent_class.get_default_cfg(), update_unknown_key_only=True)
        return cfg

    def eval(self, **kwargs):
        pass
    
    def _check_params(self):
        pass
    
    def set_callback_list(self, callbacklist: CallbackList):
        self.callback_list = callbacklist

    def set_dataloaders(self, train_loader, test_loader, valid_loader=None):
        self.train_loader = train_loader
        self.valid_loader = valid_loader
        self.test_loader = test_loader

    def add_extra_data(self, **kwargs):
        self.extra_data = kwargs




from .raw2mid import BaseRaw2Mid
from .assist_0910 import R2M_ASSIST_0910
from .frcsub import R2M_FrcSub
from .assist_1213 import R2M_ASSIST_1213
from .math1 import R2M_Math1
from .math2 import R2M_Math2
from .aaai_2023 import R2M_AAAI_2023
from .algebra2005 import R2M_Algebra_0506
from .assist_1516 import R2M_ASSIST_1516
from .assist_2017 import R2M_ASSIST_17
from .bridge2006 import R2M_Bridge2Algebra_0607
from .ednet_kt1 import R2M_EdNet_KT1
from .junyi_area_topic_as_cpt import R2M_Junyi_AreaTopicAsCpt
from .junyi_exer_as_cpt import R2M_JunyiExerAsCpt
from .nips12 import R2M_Eedi_20_T12
from .nips34 import R2M_Eedi_20_T34
from .simulated5 import R2M_Simulated5
from .slp_english import R2M_SLP_English
from .slp_math import R2M_SLP_Math

# look up api dict
_cli_api_dict_ = {}
_cli_api_dict_['R2M_ASSIST_0910'] = R2M_ASSIST_0910.from_cli
_cli_api_dict_['R2M_FrcSub'] = R2M_FrcSub.from_cli
_cli_api_dict_['R2M_ASSIST_1213'] = R2M_ASSIST_1213.from_cli
_cli_api_dict_['R2M_Math1'] = R2M_Math1.from_cli
_cli_api_dict_['R2M_Math2'] = R2M_Math2.from_cli
_cli_api_dict_['R2M_Xueersi_2023'] = R2M_AAAI_2023.from_cli
_cli_api_dict_['R2M_Algebra_0506'] = R2M_Algebra_0506.from_cli
_cli_api_dict_['R2M_ASSIST_1516'] = R2M_ASSIST_1516.from_cli
_cli_api_dict_['R2M_ASSIST_17'] = R2M_ASSIST_17.from_cli
_cli_api_dict_['R2M_Bridge2Algebra_0607'] = R2M_Bridge2Algebra_0607.from_cli
_cli_api_dict_['R2M_EdNet_KT1'] = R2M_EdNet_KT1.from_cli
_cli_api_dict_['R2M_Junyi_AreaTopicAsCpt'] = R2M_Junyi_AreaTopicAsCpt.from_cli
_cli_api_dict_['R2M_JunyiExerAsCpt'] = R2M_JunyiExerAsCpt.from_cli
_cli_api_dict_['R2M_Eedi_20_T12'] = R2M_Eedi_20_T12.from_cli
_cli_api_dict_['R2M_Eedi_20_T34'] = R2M_Eedi_20_T34.from_cli
_cli_api_dict_['R2M_Simulated5'] = R2M_Simulated5.from_cli
_cli_api_dict_['R2M_SLP_Math'] = R2M_SLP_Math.from_cli
_cli_api_dict_['R2M_SLP_English'] = R2M_SLP_English.from_cli



# -*- coding: utf-8 -*-
# @Author : Xiangzhi Chen
# @Github : kervias

from .base_traintpl import BaseTrainTPL
from .gd_traintpl import GDTrainTPL
from .general_traintpl import GeneralTrainTPL
from .atkt_traintpl import AtktTrainTPL
from .dcd_traintpl import DCDTrainTPL
from .adversarial_traintpl import AdversarialTrainTPL
from .group_cd_traintpl import GroupCDTrainTPL



r"""
DINA
##########################################

Reference:
    Jimmy De La Torre. Dina model and parameter estimation: A didactic. Journal of educational and behavioral statistics, 34(1):115–130, 2009

Reference Code:
    https://github.com/bigdata-ustc/EduCDM/blob/main/EduCDM/DINA/

"""

import torch
import torch.nn as nn
import torch.nn.functional as F
import numpy as np
from ..gd_basemodel import GDBaseModel
import torch.autograd as autograd


class DINA(GDBaseModel):
    default_cfg = {
        "step": 0,
        "max_step": 1000,
        "max_slip": 0.4,
        "max_guess": 0.4,
    }
    def __init__(self, cfg):
        super().__init__(cfg)
    
    def add_extra_data(self, **kwargs):
        self.Q_mat = kwargs['Q_mat'].to(self.device)

    def build_cfg(self):
        self.n_user = self.datatpl_cfg['dt_info']['stu_count']
        self.n_item = self.datatpl_cfg['dt_info']['exer_count']
        self.n_cpt = self.datatpl_cfg['dt_info']['cpt_count']

        self.step = 0
        self.max_step = self.modeltpl_cfg['max_step']
        self.max_slip =  self.modeltpl_cfg['max_slip']
        self.max_guess =  self.modeltpl_cfg['max_guess']
        self.emb_dim = self.n_cpt

    def build_model(self):
        self.guess = nn.Embedding(self.n_item, 1)
        self.slip = nn.Embedding(self.n_item, 1)
        self.theta = nn.Embedding(self.n_user, self.emb_dim)

    def forward(self, stu_id, exer_id, **kwargs):
        items_Q_mat = self.Q_mat[exer_id]
        theta = self.theta(stu_id)
        slip = torch.squeeze(torch.sigmoid(self.slip(exer_id)) * self.max_slip)
        guess = torch.squeeze(torch.sigmoid(self.guess(exer_id)) * self.max_guess)

        knowledge = items_Q_mat
        if self.training:
            n = torch.sum(knowledge * (torch.sigmoid(theta) - 0.5), dim=1)
            t, self.step = max((np.sin(2 * np.pi * self.step / self.max_step) + 1) / 2 * 100,
                               1e-6), self.step + 1 if self.step < self.max_step else 0
            return torch.sum(
                torch.stack([1 - slip, guess]).T * torch.softmax(torch.stack([n, torch.zeros_like(n)]).T / t, dim=-1),
                dim=1
            )
        else:
            n = torch.prod(knowledge * (theta >= 0) + (1 - knowledge), dim=1)
            return (1 - slip) ** n * guess ** (1 - n)
    
    def get_main_loss(self, **kwargs):
        stu_id = kwargs['stu_id']
        exer_id = kwargs['exer_id']
        label = kwargs['label']
        pd = self(stu_id, exer_id).flatten()
        loss = F.binary_cross_entropy(input=pd, target=label)
        return {
            'loss_main': loss
        }

    def get_loss_dict(self, **kwargs):
        return self.get_main_loss(**kwargs)
    
    @torch.no_grad()
    def predict(self, stu_id, exer_id, **kwargs):
        return {
            'y_pd': self(stu_id, exer_id).flatten(),
        }
                
    def get_stu_status(self, stu_id=None):
        if stu_id is not None:
            return self.theta(stu_id)
        else:
            return self.theta.weight

class STEFunction(autograd.Function):
    @staticmethod
    def forward(ctx, input):
        return (input > 0).float()

    @staticmethod
    def backward(ctx, grad_output):
        return F.hardtanh(grad_output)


class StraightThroughEstimator(nn.Module):
    def __init__(self):
        super(StraightThroughEstimator, self).__init__()

    def forward(self, x):
        x = STEFunction.apply(x)
        return x


class STEDINA(DINA):
    default_cfg = {
        "max_slip": 0.4,
        "max_guess": 0.4,
    }
    def __init__(self, cfg):
        super().__init__(cfg)
        self.sign = StraightThroughEstimator()

    def build_cfg(self):
        self.n_user = self.datatpl_cfg['dt_info']['stu_count']
        self.n_item = self.datatpl_cfg['dt_info']['exer_count']
        self.n_cpt = self.datatpl_cfg['dt_info']['cpt_count']

        self.max_slip =  self.modeltpl_cfg['max_slip']
        self.max_guess =  self.modeltpl_cfg['max_guess']
        self.emb_dim = self.n_cpt

    def forward(self, stu_id, exer_id):
        theta = self.theta(stu_id)
        theta = self.sign(self.theta(stu_id))
        knowledge = self.Q_mat[exer_id]
        slip = torch.squeeze(torch.sigmoid(self.slip(exer_id)) * self.max_slip)
        guess = torch.squeeze(torch.sigmoid(self.guess(exer_id)) * self.max_guess)
        mask_theta = (knowledge == 0) + (knowledge == 1) * theta
        n = torch.prod((mask_theta + 1) / 2, dim=-1)
        return torch.pow(1 - slip, n) * torch.pow(guess, 1 - n)



# -*- coding: utf-8 -*-
# @Author : Xiangzhi Chen
# @Github : kervias

import torch.nn as nn
import logging
from edustudio.utils.common import UnifyConfig
import importlib

class BaseModel(nn.Module):
    """The basic protocol for implementing a model template
    """
    default_cfg = {}
    def __init__(self, cfg):
        super().__init__()
        self.cfg: UnifyConfig = cfg
        self.datatpl_cfg: UnifyConfig = cfg.datatpl_cfg
        self.evaltpl_cfg: UnifyConfig = cfg.evaltpl_cfg
        self.traintpl_cfg: UnifyConfig = cfg.traintpl_cfg
        self.frame_cfg: UnifyConfig = cfg.frame_cfg
        self.modeltpl_cfg: UnifyConfig = cfg.modeltpl_cfg
        self.logger: logging.Logger = cfg.logger

    @classmethod
    def from_cfg(cls, cfg: UnifyConfig):
        """an interface to instantiate a model

        Args:
            cfg (UnifyConfig): the global config object

        Returns:
           BaseModel
        """
        return cls(cfg)
    
    def add_extra_data(self, **kwargs):
        """an interface to get extra data except the data from forward API
        """
        pass

    @classmethod
    def get_default_cfg(cls, **kwargs):
        """get the final default_cfg
        """
        cfg = UnifyConfig()
        for _cls in cls.__mro__:
            if not hasattr(_cls, 'default_cfg'):
                break
            cfg.update(_cls.default_cfg, update_unknown_key_only=True)
        return cfg


# class BaseProxyModel(object):
#     default_cfg = {
#         'backbone_modeltpl_cls': 'BaseModel'
#     }
    
#     @classmethod
#     def from_cfg(cls, cfg):
#         backbone_modeltpl_cls = cls.get_backbone_cls(cfg.modeltpl_cfg.backbone_modeltpl_cls)
#         cls.register_functions(backbone_cls=backbone_modeltpl_cls)
#         return backbone_modeltpl_cls.from_cfg(cfg)
    
#     @classmethod
#     def register_functions(cls, backbone_cls):
#         pass

#     @classmethod
#     def get_backbone_cls(cls, backbone_modeltpl_cls):
#         if isinstance(backbone_modeltpl_cls, str):
#             backbone_modeltpl_cls = importlib.import_module('edustudio.model').\
#                 __getattribute__(backbone_modeltpl_cls)
#         elif isinstance(backbone_modeltpl_cls, BaseModel):
#             backbone_modeltpl_cls = backbone_modeltpl_cls
#         else:
#             raise ValueError(f"Unknown type of backbone_modeltpl_cls: {backbone_modeltpl_cls}")
#         return backbone_modeltpl_cls

#     @classmethod
#     def get_default_cfg(cls, backbone_modeltpl_cls, **kwargs):
#         parent_class = cls.__base__
#         cfg = UnifyConfig(cls.default_cfg)
#         cfg.backbone_modeltpl_cls = backbone_modeltpl_cls or cfg.backbone_modeltpl_cls
#         if hasattr(parent_class, 'get_default_cfg'):
#             cfg.update(parent_class.get_default_cfg(backbone_modeltpl_cls=cfg.backbone_modeltpl_cls, **kwargs), update_unknown_key_only=True)
#         if cls is BaseProxyModel:
#             backbone_modeltpl_cls = cls.get_backbone_cls(backbone_modeltpl_cls=cfg.backbone_modeltpl_cls, **kwargs)
#             cfg.update(backbone_modeltpl_cls.get_default_cfg(**kwargs), update_unknown_key_only=True)
#         return cfg



class BaseProxyModel(object):
    """The baisc protocol to implement a proxy model template
    """
    default_cfg = {
        'backbone_modeltpl_cls': 'BaseModel'
    }
    
    @classmethod
    def from_cfg_proxy(cls, cfg):
        """an interface to instantiate a proxy model

        Args:
            cfg (UnifyConfig): the global config object

        Returns:
           BaseModel
        """
        backbone_modeltpl_cls = cls.get_backbone_cls(cfg.modeltpl_cfg.backbone_modeltpl_cls)
        new_cls = cls.get_new_cls(p_cls=backbone_modeltpl_cls)
        return new_cls.from_cfg(cfg)
    
    @classmethod
    def get_backbone_cls(cls, backbone_modeltpl_cls):
        """get backbone class from its name

        Args:
            backbone_modeltpl_cls (_type_): class name or class address

        Returns:
            BaseProxyModel: the class address of proxy model
        """
        if isinstance(backbone_modeltpl_cls, str):
            backbone_modeltpl_cls = importlib.import_module('edustudio.model').\
                __getattribute__(backbone_modeltpl_cls)
        elif issubclass(backbone_modeltpl_cls, BaseModel):
            backbone_modeltpl_cls = backbone_modeltpl_cls
        else:
            raise ValueError(f"Unknown type of backbone_modeltpl_cls: {backbone_modeltpl_cls}")
        return backbone_modeltpl_cls

    @classmethod
    def get_new_cls(cls, p_cls):
        """dynamic inheritance

        Args:
            p_cls (BaseModel): parent class

        Returns:
            BaseProxyModel: A inherited class
        """
        new_cls = type(cls.__name__ + "_proxy", (cls, p_cls), {})
        return new_cls

    @classmethod
    def get_default_cfg(cls, backbone_modeltpl_cls):
        """get the final default_cfg

        Args:
            backbone_modeltpl_cls (str/BaseModel): name or address of backbone class

        Returns:
            UnifyConfig: the final default_cfg object
        """
        bb_cls = None
        if backbone_modeltpl_cls is not None:
            bb_cls = cls.get_backbone_cls(backbone_modeltpl_cls)
        else:
            for _cls in cls.__mro__:
                if not hasattr(_cls, 'default_cfg'):
                    break
                bb_cls = _cls.default_cfg.get('backbone_modeltpl_cls', None)
                if bb_cls is not None: break
            assert bb_cls is not None
            bb_cls = cls.get_backbone_cls(bb_cls)
        
        cfg = UnifyConfig()
        cfg.backbone_modeltpl_cls = bb_cls
        cfg.backbone_modeltpl_cls_name = bb_cls.__name__
        new_cls = cls.get_new_cls(p_cls=bb_cls)
        for _cls in new_cls.__mro__:
            if not hasattr(_cls, 'default_cfg'):
                break
            cfg.update(_cls.default_cfg, update_unknown_key_only=True)
        return cfg



# -*- coding: utf-8 -*-
# @Author : Xiangzhi Chen
# @Github : kervias

from .callback import Callback
from collections import defaultdict
from typing import List, Sequence, Optional
import json
import os
import numpy as np
from edustudio.utils.common import NumpyEncoder


class History(Callback):
    def __init__(self, folder_path, exclude_metrics=(), plot_curve=False):
        super(History, self).__init__()
        self.log_as_time = {}
        self.exclude_metrics = set(exclude_metrics)
        self.folder_path = folder_path
        if not os.path.exists(self.folder_path):
            os.makedirs(self.folder_path)
            
        self.plot_curve = plot_curve
        if self.plot_curve:
            import matplotlib
            if not os.path.exists(self.folder_path+"/plots/"):
                os.makedirs(self.folder_path+"/plots/")
        self.names = None
    
    def on_epoch_end(self, epoch, logs=None, **kwargs):
        self.log_as_time[epoch] = {k:v for k,v in logs.items() if isinstance(v, (int, str, float)) and k not in self.exclude_metrics}
        if self.names is None:
            self.names = set(self.log_as_time[epoch].keys())
        else:
            assert self.names == set(self.log_as_time[epoch].keys())
        

    def on_train_end(self, logs=None, **kwargs):
        self.dump_json(self.log_as_time, os.path.join(self.folder_path, "history.json"))

        if self.plot_curve:
            self.logger.info("[CALLBACK]-History: Plot Curve...")
            self.plot()
            self.logger.info("[CALLBACK]-History: Plot Succeed!")

    def plot(self):
        import matplotlib.pyplot as plt
        epoch_num = len(self.log_as_time)
        x_arr = np.arange(1, epoch_num+1)
        for name in self.names:
            value_arr = [self.log_as_time[i][name] for i in range(1, epoch_num+1)]
            plt.figure()
            plt.title(name)
            plt.xlabel("epoch")
            plt.ylabel("value")
            plt.plot(x_arr, value_arr)
            plt.autoscale()
            plt.savefig(f"{self.folder_path}/plots/{name}.png", dpi=500, bbox_inches='tight', pad_inches=0.1)
        
    @staticmethod
    def dump_json(data, filepath):
        if not os.path.exists(os.path.dirname(filepath)):
            os.makedirs(os.path.dirname(filepath))
        with open(filepath, 'w', encoding='utf8') as f:
            json.dump(data, fp=f, indent=2, ensure_ascii=False, cls=NumpyEncoder)



from .dkt import DKT
from .dkvmn import DKVMN
from .dkt_plus import DKT_plus
from .iekt import IEKT
from .dkt_forget import DKTForget
from .akt import AKT
from .ckt import CKT
from .hawkeskt import HawkesKT
from .kqn import KQN
from .deep_irt import DeepIRT
from .sakt import SAKT
from .dkt_dsc import DKTDSC
from .lpkt import LPKT
from .simplekt import SimpleKT
from .saint import SAINT
from .skvmn import SKVMN
from .saint_plus import SAINT_plus
from .ct_ncm import  CT_NCM
from .lpkt_s import LPKT_S
from .rkt import RKT
from .qikt import QIKT
from .dtransformer import DTransformer
from .qdkt import QDKT
from .cl4kt import CL4KT
from .dimkt import DIMKT
from .GKT import *
from .atkt import ATKT
from .eernn import EERNNM, EERNNA
from .ekt import EKTM, EKTA



# -*- coding: utf-8 -*-
# @Author : Xiangzhi Chen
# @Github : kervias

import logging
import importlib
from edustudio.utils.common import UnifyConfig
from edustudio.datatpl import BaseDataTPL, BaseProxyDataTPL
from edustudio.model import BaseModel, BaseProxyModel
from edustudio.evaltpl import BaseEvalTPL
from edustudio.datatpl import BaseDataTPL
from typing import Sequence
from edustudio.utils.common import set_same_seeds


class BaseTrainTPL(object):
    """The basic protocol for implementing a training template
    """

    default_cfg = {
        "seed": 2023
    }

    def __init__(self, cfg: UnifyConfig):
        self.cfg: UnifyConfig = cfg
        self.datatpl_cfg: UnifyConfig = cfg.datatpl_cfg
        self.evaltpl_cfg: UnifyConfig = cfg.evaltpl_cfg
        self.traintpl_cfg: UnifyConfig = cfg.traintpl_cfg
        self.frame_cfg: UnifyConfig = cfg.frame_cfg
        self.modeltpl_cfg: UnifyConfig = cfg.modeltpl_cfg
        self.logger: logging.Logger = logging.getLogger("edustudio")
        self._check_params()

        if isinstance(self.datatpl_cfg['cls'], str):
            self.datatpl_cls: BaseDataTPL = importlib.import_module('edustudio.datatpl').\
                __getattribute__(self.datatpl_cfg['cls'])
        else:
            self.datatpl_cls = self.datatpl_cfg['cls']
        
        if isinstance(self.modeltpl_cfg['cls'], str):
            self.model_cls: BaseModel = importlib.import_module('edustudio.model').\
                __getattribute__(self.modeltpl_cfg['cls'])
        else:
            self.model_cls = self.modeltpl_cfg['cls']
        
        self.evaltpl_clses: Sequence[BaseEvalTPL] = [
            importlib.import_module('edustudio.evaltpl').__getattribute__(tpl) if isinstance(tpl, str) else tpl
            for tpl in self.evaltpl_cfg['clses'] 
        ]

        set_same_seeds(self.datatpl_cfg['seed'])

        self.model = self.get_model_obj()
        self.datatpl = self.get_datatpl_obj()
        self.evaltpls = self.get_evaltpl_obj_list()

    @classmethod
    def get_default_cfg(cls):
        parent_class = cls.__base__
        cfg = UnifyConfig(cls.default_cfg)
        if hasattr(parent_class, 'get_default_cfg'):
            cfg.update(parent_class.get_default_cfg(), update_unknown_key_only=True)
        return cfg
    
    def start(self):
        """entrypoint of starting a training process
        """
        self.logger.info(f"TrainTPL {self.__class__} Started!")
        set_same_seeds(self.traintpl_cfg['seed'])

    def _check_params(self):
        """check validation of default config
        """
        pass

    def get_model_obj(self):
        if issubclass(self.model_cls, BaseModel):
            model: BaseModel = self.model_cls.from_cfg(self.cfg)
        elif issubclass(self.model_cls, BaseProxyModel):
            model: BaseModel = self.model_cls.from_cfg_proxy(self.cfg)
        else:
            raise ValueError(f"unsupported model_cls: {self.model_cls}")
        return model

    def get_datatpl_obj(self):
        if issubclass(self.datatpl_cls, BaseDataTPL):
            datatpl: BaseDataTPL = self.datatpl_cls.from_cfg(self.cfg)
        elif issubclass(self.datatpl_cls, BaseProxyDataTPL):
            datatpl: BaseDataTPL = self.datatpl_cls.from_cfg_proxy(self.cfg)
        else:
            raise ValueError(f"unsupported datatpl_cls: {self.datatpl_cls}")
        return datatpl

    def get_evaltpl_obj_list(self):
        evaltpls: Sequence[BaseEvalTPL] = [cls(self.cfg) for cls in self.evaltpl_clses]
        return evaltpls




from .raw2mid import BaseRaw2Mid
import pandas as pd

r"""
R2M_ASSIST_0910
#####################################
ASSIST_0910 dataset preprocess
"""


class R2M_ASSIST_0910(BaseRaw2Mid):
    """R2M_ASSIST_0910 is to preprocess ASSISTment 2009-2010 dataset"""

    def process(self):
        super().process()
        pd.set_option("mode.chained_assignment", None)  # ingore warning
        df = pd.read_csv(f"{self.rawpath}/skill_builder_data.csv", encoding='utf-8', low_memory=False)
        # 去除空值
        df = df.dropna(subset=['skill_id'])

        # 学校和班级合并
        df['class_id'] = df['school_id'].astype(str) + '_' + df['student_class_id'].astype(str)
        # print(df)

        # 进行映射
        df = df[['user_id', 'assignment_id', 'problem_id', 'correct', 'ms_first_response', 'overlap_time', 'class_id',
                 'skill_id']]

        def sort(data, column):
            '''将原始数据对指定列进行排序，并完成0-num-1映射'''

            # 非 class_id、非ms_first_response单独处理
            if (column != 'class_id'):
                data.sort_values(column, inplace=True)

            # 处理 class_id
            elif (column != 'ms_first_response'):
                data[['school_id', 'group_id']] = data['class_id'].str.split('_', expand=True)
                data['school_id'] = data['school_id'].astype(int)
                data['group_id'] = data['group_id'].astype(int)
                data.sort_values(['school_id', 'group_id'], inplace=True)

            # 映射非 ms_first_response 列
            if (column != 'ms_first_response'):
                value_mapping = {}
                new_value = 0
                for value in data[column].unique():
                    value_mapping[value] = new_value
                    new_value += 1

                # 映射 生成新列
                new_column = f'new_{column}'
                data[new_column] = data[column].map(value_mapping)
                del data[column]

            # 处理ms_first_response，每个同学答题顺序单独映射
            else:
                data = data.sort_values(by=['new_user_id', 'ms_first_response'])

                # 创建空字典存储同学的作答记录时间编号映射
                user_mapping = {}
                user_count = {}

                def generate_mapping(row):
                    '''生成作答记录时间编号映射的函数'''
                    user_id = row['new_user_id']
                    timestamp = row['ms_first_response']
                    if user_id not in user_mapping:
                        user_mapping[user_id] = {}
                        user_count[user_id] = 0
                    if timestamp not in user_mapping[user_id]:
                        user_mapping[user_id][timestamp] = user_count[user_id]
                        user_count[user_id] += 1
                    return user_mapping[user_id][timestamp]

                # 映射，并生成新列
                data['new_order_id'] = data.apply(generate_mapping, axis=1)

            return data

        # 处理各列
        df = sort(df, 'user_id')
        df = sort(df, 'assignment_id')
        df = sort(df, 'problem_id')
        df = sort(df, 'skill_id')
        df = sort(df, 'class_id')
        df = sort(df, 'ms_first_response')
        del df['school_id']
        del df['group_id']
        # print(df)

        # 修改列名
        df = df.rename(
            columns={'new_user_id': 'stu_id:token', 'new_problem_id': 'exer_id:token', 'correct': 'label:float',
                     'ms_first_response': 'start_timestamp:float', 'new_class_id': 'class_id:token',
                     'new_skill_id': 'cpt_seq:token_seq', 'overlap_time': 'cost_time:float',
                     'new_assignment_id': 'assignment_id:token_seq', 'new_order_id': 'order_id:float'})
        # 指定列的新顺序
        new_column_order = ['stu_id:token', 'exer_id:token', 'label:float', 'start_timestamp:float', 'cost_time:float',
                            'order_id:float',
                            'class_id:token', 'cpt_seq:token_seq', 'assignment_id:token_seq']
        df = df.reindex(columns=new_column_order)
        # print(df)

        # df_inter 相关处理
        df_inter = df[['stu_id:token', 'exer_id:token', 'label:float', 'start_timestamp:float', 'cost_time:float',
                       'order_id:float']]
        df_inter.drop_duplicates(inplace=True)
        df_inter.sort_values('stu_id:token', inplace=True)
        # print(df_inter)

        # df_user 相关处理
        df_user = df[['stu_id:token', 'class_id:token']]
        df_user.drop_duplicates(inplace=True)
        df_user.sort_values('stu_id:token', inplace=True)
        # print(df_user)

        # df_exer 相关处理

        # 处理列名
        df_exer = df[['exer_id:token', 'cpt_seq:token_seq', 'assignment_id:token_seq']]
        df_exer.sort_values(by='exer_id:token', inplace=True)
        df_exer.drop_duplicates(inplace=True)

        # 合并 cpt_seq
        grouped_skills = df_exer[['exer_id:token', 'cpt_seq:token_seq']]
        grouped_skills.drop_duplicates(inplace=True)
        grouped_skills.sort_values(by='exer_id:token', inplace=True)
        grouped_skills['exer_id:token'] = grouped_skills['exer_id:token'].astype(str)
        grouped_skills['cpt_seq:token_seq'] = grouped_skills['cpt_seq:token_seq'].astype(str)
        grouped_skills = grouped_skills.groupby('exer_id:token')['cpt_seq:token_seq'].agg(','.join).reset_index()

        # 合并 assignment_id
        grouped_assignments = df_exer[['exer_id:token', 'assignment_id:token_seq']]
        grouped_assignments.drop_duplicates(inplace=True)
        grouped_assignments.sort_values(by='assignment_id:token_seq', inplace=True)
        grouped_assignments['exer_id:token'] = grouped_assignments['exer_id:token'].astype(str)
        grouped_assignments['assignment_id:token_seq'] = grouped_assignments['assignment_id:token_seq'].astype(str)
        grouped_assignments = grouped_assignments.groupby('exer_id:token')['assignment_id:token_seq'].agg(
            ','.join).reset_index()

        # 合并结果
        df_exer = pd.merge(grouped_skills, grouped_assignments, on='exer_id:token', how='left')
        df_exer['exer_id:token'] = df_exer['exer_id:token'].astype(int)
        df_exer.sort_values(by='exer_id:token', inplace=True)
        # print(df_exer)

        # # Save MidData

        df_inter.to_csv(f"{self.midpath}/{self.dt}.inter.csv", index=False, encoding='utf-8')
        df_user.to_csv(f"{self.midpath}/{self.dt}.stu.csv", index=False, encoding='utf-8')
        df_exer.to_csv(f"{self.midpath}/{self.dt}.exer.csv", index=False, encoding='utf-8')
        pd.set_option("mode.chained_assignment", "warn")  # igore warning


from ..common import GeneralDataTPL

class KTInterDataTPL(GeneralDataTPL):
    default_cfg = {
        'mid2cache_op_seq': ['M2C_Label2Int', 'M2C_ReMapId', 'M2C_BuildSeqInterFeats', 'M2C_RandomDataSplit4KT'],
    }




from ..common import EduDataTPL


class CDGKDataTPL(EduDataTPL):
    default_cfg = {
        'mid2cache_op_seq': ['M2C_Label2Int', 'M2C_FilterRecords4CD', 'M2C_ReMapId', 'M2C_RandomDataSplit4CD', 'M2C_GenQMat', 'M2C_CDGK_OP'],
    }

    # @classmethod
    # def load_data(cls, cfg): # 只在middata存在时调用
    #     kwargs = super().load_data(cfg)
    #     if cfg.datatpl_cfg['has_cpt2group_file'] is True:
    #         new_kwargs = cls.load_cpt2group(cfg)
    #         for df in new_kwargs.values(): 
    #             if df is not None:
    #                 cls._preprocess_feat(df) # 类型转换
    #         kwargs.update(new_kwargs)
    #     else:
    #         kwargs['df_cpt2group'] = None
    #     return kwargs
    

    # @classmethod
    # def load_cpt2group(cls, cfg):
    #     cpt2group_fph = f'{cfg.frame_cfg.data_folder_path}/middata/{cfg.dataset}.cpt2group.csv'
    #     sep = cfg.datatpl_cfg['seperator']
    #     df_cpt2group = cls._load_atomic_csv(cpt2group_fph, sep=sep)
    #     return {"df_cpt2group": df_cpt2group}



import pandas as pd
from .raw2mid import BaseRaw2Mid

r"""
R2M_Junyi_Area_Topic_As_Cpt
########################
"""


class R2M_Junyi_AreaTopicAsCpt(BaseRaw2Mid):
    """R2M_Junyi_Area_Topic_As_Cpt is a class used to handle the Junyi dataset, where we consider the exercise's KC(Area) and KC(Topic) as the basis for constructing the cpt_seq (concept sequence)."""

    def process(self):
        super().process()
        pd.set_option("mode.chained_assignment", None)  # ignore warning
        # 读取数据集，并展示相关数据
        raw_data = pd.read_table(f"{self.rawpath}/junyi_ProblemLog_for_PSLC.txt", sep='\t', encoding='utf-8',
                                 low_memory=True)
        data = raw_data.head()
        # 去除提示的交互
        data = data.drop(data[data.Outcome == 'HINT'].index)
        data = data[
            ['Anon Student Id', 'Time', 'Level (Section)', 'Problem Name', 'Problem Start Time', 'Outcome',
             'KC (Exercise)',
             'KC (Topic)', 'KC (Area)']]
        data = data.sort_values(by='Time', ascending=True)
        data['order:token'] = range(len(data))
        data['cost_time:float'] = data['Time'] - data['Problem Start Time']

        inter = data.rename(
            columns={'Anon Student Id': 'stu_id:token', 'Problem Name': 'exer_name:token', 'Outcome': 'label:float',
                     'Problem Start Time': 'start_timestamp:float', }) \
            .reindex(
            columns=['stu_id:token', 'exer_name:token', 'label:float', 'start_timestamp:float', 'cost_time:float',
                     'order:token'])

        # 定义一个name_to_id的函数，返回值为两个dict,一个是name2id，一个是id2name
        # 主要用于将唯一标识的字符串生成自增的id
        def name_to_id(l: list()):
            if l is None:
                return None, None
            l = list(l)
            name2id = dict(zip(l, range(len(l))))
            id2name = dict(zip(range(len(l)), l))
            return name2id, id2name

        # 进行exer_name 转 exer_id
        unique_exer = list(inter['exer_name:token'].unique())
        exer2id, id2exer = name_to_id(unique_exer)
        inter['exer_id:token'] = inter['exer_name:token'].map(exer2id)
        # 确定label
        str2label = {'CORRECT': 1, 'INCORRECT': 0}
        inter['label:float'] = inter['label:float'].map(str2label)
        df_inter = inter[
            ['stu_id:token', 'exer_id:token', 'label:float', 'start_timestamp:float', 'cost_time:float', 'order:token']]

        # 处理用户信息
        df_stu = data['Anon Student Id'].unique()
        # df_stu['class_id'] = None
        # df_stu['gender'] = None

        # 处理习题信息
        exer = data[['Problem Name', 'Level (Section)', 'KC (Exercise)', 'KC (Area)', 'KC (Topic)']].copy()
        # 处理习题，先去重
        exer = exer.drop_duplicates(subset=['Problem Name'], keep='first')
        exer['exer_id'] = exer['Problem Name'].map(exer2id)
        # 处理assignment name to id的映射
        unique_assignment = list(exer['Level (Section)'].unique())
        assignment2id, id2assignment = name_to_id(unique_assignment)
        exer['assignment_id'] = exer['Level (Section)'].map(assignment2id)
        # 获取KCs
        exer['KC (Area)'] = exer['KC (Area)'] + '&Area'  # 标明知识点来自Area
        exer['KC (Topic)'] = exer['KC (Topic)'] + '&Topic'  # 标明知识点来自Topic
        unique_kc_area = set(exer['KC (Area)'])
        unique_kc_topic = set(exer['KC (Topic)'])
        # 基于Area和Topic字段生成所有KC,及kc2id,id2kc
        unique_kc = list(unique_kc_area | unique_kc_topic)
        kc2id, id2kc = name_to_id(unique_kc)
        exer['cpt_seq_area'] = exer['KC (Area)'].map(lambda x: kc2id[x])
        exer['cpt_seq_topic'] = exer['KC (Topic)'].map(lambda x: kc2id[x])
        exer['cpt_seq'] = '[' + exer['cpt_seq_area'].map(str) + ',' + exer['cpt_seq_topic'].map(str) + ']'

        df_exer = exer.rename(columns={
            'exer_id': 'exer_id:token', 'assignment_id': 'assignment_id:token_seq',
            'cpt_seq': 'cpt_seq:token_seq',
            'cpt_seq_area': 'cpt_seq_level_1:token_seq', 'cpt_seq_topic': 'cpt_seq_level_2:token_seq'
            # area - level1;    topic - level2
        }).reindex(
            columns=['exer_id:token', 'assignment_id:token_seq', 'cpt_seq:token_seq', 'cpt_seq_level_1:token_seq',
                     'cpt_seq_level_2:token_seq'])

        # 处理cpt_tree信息
        # topic ∈ area
        topic_area = exer.drop_duplicates(subset=['cpt_seq_topic'], inplace=False)[['cpt_seq_topic', 'cpt_seq_area']]
        # cpt_id:token    parent_id:token    level:float
        cpt_data = []
        # topic, area, level = 2
        topic_area.apply(lambda x: cpt_data.append([x['cpt_seq_topic'], x['cpt_seq_area'], 2]), axis=1)
        area = topic_area.drop_duplicates(subset=['cpt_seq_area'], inplace=False)[['cpt_seq_area']]
        # area, none, level = 1
        area.apply(lambda x: cpt_data.append([x['cpt_seq_area'], None, 1]), axis=1)
        df_cpt = pd.DataFrame(columns=['cpt_id:token', 'parent_id:token', 'level:float'], data=cpt_data)
        df_cpt.sort_values(by='level:float', ascending=True, inplace=True)

        # 此处将数据保存到`self.midpath`中

        df_inter.to_csv(f"{self.midpath}/{self.dt}.inter.csv", index=False, encoding='utf-8')
        # df_stu.to_csv(f"{self.midpath}/{self.dt}.stu.csv", index=False, encoding='utf-8')
        df_exer.to_csv(f"{self.midpath}/{self.dt}.exer.csv", index=False, encoding='utf-8')
        df_cpt.to_csv(f"{self.midpath}/{self.dt}.cpt_tree.csv", index=False, encoding='utf-8')
        pd.set_option("mode.chained_assignment", "warn")  # ignore warning
        return



r"""
DKT
##########################################

Reference:
    Chris Piech et al. "Deep knowledge tracing" in NIPS 2015.

"""

from ..gd_basemodel import GDBaseModel
import torch.nn as nn
import torch
import torch.nn.functional as F


class DKT(GDBaseModel):
    default_cfg = {
        'emb_size': 100,
        'hidden_size': 100,
        'num_layers': 1,
        'dropout_rate': 0.2,
        'rnn_or_lstm': 'lstm',
    }

    def __init__(self, cfg):
        super().__init__(cfg)

    def build_cfg(self):
        self.n_user = self.datatpl_cfg['dt_info']['stu_count']
        self.n_item = self.datatpl_cfg['dt_info']['exer_count']
        assert self.modeltpl_cfg['rnn_or_lstm'] in {'rnn', 'lstm'}

    def build_model(self):
        self.exer_emb = nn.Embedding(
            self.n_item * 2, self.modeltpl_cfg['emb_size']
        )
        if self.modeltpl_cfg['rnn_or_lstm'] == 'rnn':
            self.seq_model = nn.RNN(
                self.modeltpl_cfg['emb_size'], self.modeltpl_cfg['hidden_size'], 
                self.modeltpl_cfg['num_layers'], batch_first=True
            )
        else:
            self.seq_model = nn.LSTM(
                self.modeltpl_cfg['emb_size'], self.modeltpl_cfg['hidden_size'], 
                self.modeltpl_cfg['num_layers'], batch_first=True
            )
        self.dropout_layer = nn.Dropout(self.modeltpl_cfg['dropout_rate'])
        self.fc_layer = nn.Linear(self.modeltpl_cfg['hidden_size'], self.n_item)

    def forward(self, exer_seq, label_seq, **kwargs):
        input_x = self.exer_emb(exer_seq + label_seq.long() * self.n_item)
        output, _ = self.seq_model(input_x)
        output = self.dropout_layer(output)
        y_pd = self.fc_layer(output).sigmoid()
        return y_pd
    
    @torch.no_grad()
    def predict(self, **kwargs):
        y_pd = self(**kwargs)
        y_pd = y_pd[:, :-1].gather(
            index=kwargs['exer_seq'][:, 1:].unsqueeze(dim=-1), dim=2
        ).squeeze(dim=-1)
        y_pd = y_pd[kwargs['mask_seq'][:, 1:] == 1]
        y_gt = None
        if kwargs.get('label_seq', None) is not None:
            y_gt = kwargs['label_seq'][:, 1:]
            y_gt = y_gt[kwargs['mask_seq'][:, 1:] == 1]
        return {
            'y_pd': y_pd,
            'y_gt': y_gt
        }

    def get_main_loss(self, **kwargs):
        y_pd = self(**kwargs)
        y_pd = y_pd[:, :-1].gather(
            index=kwargs['exer_seq'][:, 1:].unsqueeze(dim=-1), dim=2
        ).squeeze(dim=-1)
        y_pd = y_pd[kwargs['mask_seq'][:, 1:] == 1]
        y_gt = kwargs['label_seq'][:, 1:]
        y_gt = y_gt[kwargs['mask_seq'][:, 1:] == 1]
        loss = F.binary_cross_entropy(
            input=y_pd, target=y_gt
        )
        return {
            'loss_main': loss
        }

    def get_loss_dict(self, **kwargs):
        return self.get_main_loss(**kwargs)



from contextlib import closing
from tqdm import tqdm
import requests
import zipfile


class BigfileDownloader(object):
    @staticmethod
    def download(url, title, filepath, chunk_size=10240):
        with closing(requests.get(url, stream=True, allow_redirects=True)) as resp:
            if resp.status_code != 200:
                raise Exception("[ERROR]: {} - {} -{}".format(str(resp.status_code), title, url))
            chunk_size = chunk_size
            content_size = int(resp.headers['content-length'])  
            with tqdm(total=content_size, desc=title, ncols=100) as pbar:
                with open(filepath, 'wb') as f:
                    for data in resp.iter_content(chunk_size=chunk_size):
                        f.write(data)
                        pbar.update(len(data))


class DecompressionUtil(object):
    @staticmethod
    def unzip_file(zip_src, dst_dir):
        r = zipfile.is_zipfile(zip_src)
        if r:     
            fz = zipfile.ZipFile(zip_src, 'r')
            for file in tqdm(fz.namelist(), desc='unzip...', ncols=100):
                fz.extract(file, dst_dir)       
        else:
            raise Exception(f'{zip_src} is not a zip file')



import torch.nn as nn
from torch.nn.init import xavier_normal_, xavier_uniform_, kaiming_normal_, kaiming_uniform_, constant_


def xavier_normal_initialization(module):
    r""" using `xavier_normal_`_ in PyTorch to initialize the parameters in
    nn.Embedding and nn.Linear layers. For bias in nn.Linear layers,
    using constant 0 to initialize.
    .. _`xavier_normal_`:
        https://pytorch.org/docs/stable/nn.init.html?highlight=xavier_normal_#torch.nn.init.xavier_normal_
    Examples:
        >>> self.apply(xavier_normal_initialization)
    """
    if isinstance(module, nn.Embedding):
        xavier_normal_(module.weight.data)
    elif isinstance(module, nn.Linear):
        xavier_normal_(module.weight.data)
        if module.bias is not None:
            constant_(module.bias.data, 0)


def xavier_uniform_initialization(module):
    r""" using `xavier_uniform_`_ in PyTorch to initialize the parameters in
    nn.Embedding and nn.Linear layers. For bias in nn.Linear layers,
    using constant 0 to initialize.
    .. _`xavier_uniform_`:
        https://pytorch.org/docs/stable/nn.init.html?highlight=xavier_uniform_#torch.nn.init.xavier_uniform_
    Examples:
        >>> self.apply(xavier_uniform_initialization)
    """
    if isinstance(module, nn.Embedding):
        xavier_uniform_(module.weight.data)
    elif isinstance(module, nn.Linear):
        xavier_uniform_(module.weight.data)
        if module.bias is not None:
            constant_(module.bias.data, 0)


def kaiming_normal_initialization(module):
    r""" using `kaiming_normal_`_ in PyTorch to initialize the parameters in
    nn.Embedding and nn.Linear layers. For bias in nn.Linear layers,
    using constant 0 to initialize.
    .. _`kaiming_normal`:
        https://pytorch.org/docs/stable/nn.init.html?highlight=kaiming_normal_#torch.nn.init.kaiming_normal_
    Examples:
        >>> self.apply(kaiming_normal_initialization)
    """
    if isinstance(module, nn.Embedding):
        kaiming_normal_(module.weight.data)
    elif isinstance(module, nn.Linear):
        kaiming_normal_(module.weight.data)
        if module.bias is not None:
            constant_(module.bias.data, 0)

def kaiming_uniform_initialization(module):
    r""" using `kaiming_uniform_`_ in PyTorch to initialize the parameters in
    nn.Embedding and nn.Linear layers. For bias in nn.Linear layers,
    using constant 0 to initialize.
    .. _`kaiming_uniform`:
        https://pytorch.org/docs/stable/nn.init.html?highlight=kaiming_uniform_#torch.nn.init.kaiming_uniform_
    Examples:
        >>> self.apply(kaiming_uniform_initialization)
    """
    if isinstance(module, nn.Embedding):
        kaiming_uniform_(module.weight.data)
    elif isinstance(module, nn.Linear):
        kaiming_uniform_(module.weight.data)
        if module.bias is not None:
            constant_(module.bias.data, 0)
